# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import re
import typing
import warnings

import kubernetes.client

import mlrun.common.schemas
import mlrun.errors
import mlrun.utils.regex

from .config import config as mlconfig

_running_inside_kubernetes_cluster = None

K8sObj = typing.Union[kubernetes.client.V1Affinity, kubernetes.client.V1Toleration]
SanitizedK8sObj = dict[str, typing.Any]
K8sObjList = typing.Union[list[K8sObj], list[SanitizedK8sObj]]


def is_running_inside_kubernetes_cluster():
    global _running_inside_kubernetes_cluster
    if _running_inside_kubernetes_cluster is None:
        try:
            kubernetes.config.load_incluster_config()
            _running_inside_kubernetes_cluster = True
        except kubernetes.config.ConfigException:
            _running_inside_kubernetes_cluster = False
    return _running_inside_kubernetes_cluster


def generate_preemptible_node_selector_requirements(
    node_selector_operator: str,
) -> list[kubernetes.client.V1NodeSelectorRequirement]:
    """
    Generate node selector requirements based on the pre-configured node selector of the preemptible nodes.
    node selector operator represents a key's relationship to a set of values.
    Valid operators are listed in :py:class:`~mlrun.common.schemas.NodeSelectorOperator`
    :param node_selector_operator: The operator of V1NodeSelectorRequirement
    :return: List[V1NodeSelectorRequirement]
    """
    match_expressions = []
    for (
        node_selector_key,
        node_selector_value,
    ) in mlconfig.get_preemptible_node_selector().items():
        match_expressions.append(
            kubernetes.client.V1NodeSelectorRequirement(
                key=node_selector_key,
                operator=node_selector_operator,
                values=[node_selector_value],
            )
        )
    return match_expressions


def generate_preemptible_nodes_anti_affinity_terms() -> (
    list[kubernetes.client.V1NodeSelectorTerm]
):
    """
    Generate node selector term containing anti-affinity expressions based on the
    pre-configured node selector of the preemptible nodes.
    Use for purpose of scheduling on node only if all match_expressions are satisfied.
    This function uses a single term with potentially multiple expressions to ensure anti affinity.
    https://kubernetes.io/docs/concepts/scheduling-eviction/assign-pod-node/#affinity-and-anti-affinity
    :return: List contains one nodeSelectorTerm with multiple expressions.
    """
    # compile affinities with operator NotIn to make sure pods are not running on preemptible nodes.
    node_selector_requirements = generate_preemptible_node_selector_requirements(
        mlrun.common.schemas.NodeSelectorOperator.node_selector_op_not_in.value
    )
    return [
        kubernetes.client.V1NodeSelectorTerm(
            match_expressions=node_selector_requirements,
        )
    ]


def generate_preemptible_nodes_affinity_terms() -> (
    list[kubernetes.client.V1NodeSelectorTerm]
):
    """
    Use for purpose of scheduling on node having at least one of the node selectors.
    When specifying multiple nodeSelectorTerms associated with nodeAffinity types,
    then the pod can be scheduled onto a node if at least one of the nodeSelectorTerms can be satisfied.
    :return: List of nodeSelectorTerms associated with the preemptible nodes.
    """
    node_selector_terms = []

    # compile affinities with operator In so pods could schedule on at least one of the preemptible nodes.
    node_selector_requirements = generate_preemptible_node_selector_requirements(
        mlrun.common.schemas.NodeSelectorOperator.node_selector_op_in.value
    )
    for expression in node_selector_requirements:
        node_selector_terms.append(
            kubernetes.client.V1NodeSelectorTerm(match_expressions=[expression])
        )
    return node_selector_terms


def generate_preemptible_tolerations() -> list[kubernetes.client.V1Toleration]:
    tolerations = mlconfig.get_preemptible_tolerations()

    toleration_objects = []
    for toleration in tolerations:
        toleration_objects.append(
            kubernetes.client.V1Toleration(
                effect=toleration.get("effect", None),
                key=toleration.get("key", None),
                value=toleration.get("value", None),
                operator=toleration.get("operator", None),
                toleration_seconds=toleration.get("toleration_seconds", None)
                or toleration.get("tolerationSeconds", None),
            )
        )
    return toleration_objects


def sanitize_label_value(value: str) -> str:
    """
    Kubernetes label values must be sanitized before they're sent to the API
    Refer to https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#syntax-and-character-set

    :param value: arbitrary string that needs to sanitized for usage on k8s labels
    :return:      string fully compliant with k8s label value expectations
    """
    return re.sub(r"([^a-zA-Z0-9_.-]|^[^a-zA-Z0-9]|[^a-zA-Z0-9]$)", "-", value[:63])


def verify_label_key(key: str, allow_k8s_prefix: bool = False):
    """
    Verify that the label key is valid for Kubernetes.
    Refer to https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#syntax-and-character-set
    """
    if not key:
        raise mlrun.errors.MLRunInvalidArgumentError("label key cannot be empty")

    prefix = ""
    parts = key.split("/")
    if len(parts) == 1:
        name = parts[0]
    elif len(parts) == 2:
        prefix, name = parts
        if len(name) == 0:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Label key name cannot be empty when a prefix is set"
            )
        if len(prefix) == 0:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Label key prefix cannot be empty"
            )

        # prefix must adhere dns_1123_subdomain
        mlrun.utils.helpers.verify_field_regex(
            f"Project.metadata.labels.'{key}'",
            prefix,
            mlrun.utils.regex.dns_1123_subdomain,
        )
    else:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "Label key can only contain one '/'"
        )

    mlrun.utils.helpers.verify_field_regex(
        f"project.metadata.labels.'{key}'",
        name,
        mlrun.utils.regex.k8s_character_limit,
    )
    mlrun.utils.helpers.verify_field_regex(
        f"project.metadata.labels.'{key}'",
        name,
        mlrun.utils.regex.qualified_name,
    )

    # Allow the use of Kubernetes reserved prefixes ('k8s.io/' or 'kubernetes.io/')
    # only when setting node selectors, not when adding new labels.
    if not allow_k8s_prefix and prefix in {"k8s.io", "kubernetes.io"}:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "Labels cannot start with 'k8s.io/' or 'kubernetes.io/'"
        )


def verify_label_value(value, label_key):
    mlrun.utils.helpers.verify_field_regex(
        f"project.metadata.labels.'{label_key}'",
        value,
        mlrun.utils.regex.label_value,
    )


def validate_node_selectors(
    node_selectors: dict[str, str], raise_on_error: bool = True
) -> bool:
    """
    Ensures that user-defined node selectors adhere to Kubernetes label standards:
    - Validates that each key conforms to Kubernetes naming conventions, with specific rules for name and prefix.
    - Ensures values comply with Kubernetes label value rules.
    - If raise_on_error is True, raises errors for invalid selectors.
    - If raise_on_error is False, logs warnings for invalid selectors.
    """

    # Helper function for handling errors or warnings
    def handle_invalid(message):
        if raise_on_error:
            raise
        else:
            warnings.warn(
                f"{message}\n"
                f"The node selector you’ve set does not meet the validation rules for the current Kubernetes version. "
                f"Please note that invalid node selectors may cause issues with function scheduling."
            )

    node_selectors = node_selectors or {}
    for key, value in node_selectors.items():
        try:
            verify_label_key(key, allow_k8s_prefix=True)
            verify_label_value(value, label_key=key)
        except mlrun.errors.MLRunInvalidArgumentError as err:
            # An error or warning is raised by handle_invalid due to validation failure.
            # Returning False indicates validation failed, allowing us to exit the function.
            handle_invalid(str(err))
            return False
    return True


def sanitize_k8s_objects(
    k8s_objects: typing.Union[None, K8sObjList, SanitizedK8sObj, K8sObj],
) -> typing.Union[list[SanitizedK8sObj], SanitizedK8sObj]:
    """Convert K8s objects to dicts. Handles single objects or lists."""
    api_client = kubernetes.client.ApiClient()
    if not k8s_objects:
        return k8s_objects

    def _sanitize_k8s_object(k8s_obj):
        return (
            api_client.sanitize_for_serialization(k8s_obj)
            if hasattr(k8s_obj, "to_dict")
            else k8s_obj
        )

    return (
        [_sanitize_k8s_object(k8s_obj) for k8s_obj in k8s_objects]
        if isinstance(k8s_objects, list)
        else _sanitize_k8s_object(k8s_objects)
    )


def sanitize_scheduling_configuration(
    tolerations: typing.Optional[list[kubernetes.client.V1Toleration]] = None,
    affinity: typing.Optional[kubernetes.client.V1Affinity] = None,
) -> tuple[
    typing.Optional[list[dict]],
    typing.Optional[dict],
]:
    """
    Sanitizes pod scheduling configuration for serialization.

    Takes affinity and tolerations and converts them to
    JSON-serializable dictionaries using the Kubernetes API client's
    sanitization method.

    Args:
        affinity: Pod affinity/anti-affinity rules
        tolerations: List of toleration rules

    Returns:
        Tuple of (sanitized_affinity, sanitized_tolerations)
        - affinity: Sanitized dict representation or None
        - tolerations: List of sanitized dict representations or None
    """
    return sanitize_k8s_objects(tolerations), sanitize_k8s_objects(affinity)


def enrich_preemption_mode(
    preemption_mode: typing.Optional[str],
    node_selector: dict[str, str],
    tolerations: list[kubernetes.client.V1Toleration],
    affinity: typing.Optional[kubernetes.client.V1Affinity],
) -> tuple[
    dict[str, str],
    list[kubernetes.client.V1Toleration],
    typing.Optional[kubernetes.client.V1Affinity],
]:
    """
    Enriches a pod spec's scheduling configuration (node selector, tolerations, affinity)
    based on the provided preemption mode.

    If no preemptible node configuration is defined in the system, or the mode is `none`,
    the original values are returned unchanged.

    Modes:
        - allow: Adds tolerations, removes preemption constraints.
        - constrain: Requires preemptible node affinity and adds tolerations.
        - prevent: Enforces scheduling on non-preemptible nodes using taints or anti-affinity.
        - none: No enrichment is applied.
    """
    if (
        not mlconfig.is_preemption_nodes_configured()
        or preemption_mode == mlrun.common.schemas.PreemptionModes.none.value
    ):
        return node_selector, tolerations, affinity

    if not preemption_mode:
        preemption_mode = mlconfig.function_defaults.preemption_mode
        mlrun.utils.logger.debug(
            "No preemption mode provided, using default",
            default_preemption_mode=preemption_mode,
        )

    enriched_node_selector = copy.deepcopy(node_selector or {})
    enriched_tolerations = _safe_copy_tolerations(tolerations or [])
    enriched_affinity = _safe_copy_affinity(affinity)
    preemptible_tolerations = generate_preemptible_tolerations()

    if handler := _get_mode_handler(preemption_mode):
        enriched_node_selector, enriched_tolerations, enriched_affinity = handler(
            enriched_node_selector,
            enriched_tolerations,
            enriched_affinity,
            preemptible_tolerations,
        )

    return (
        enriched_node_selector,
        enriched_tolerations,
        _prune_empty_affinity(enriched_affinity),
    )


def _safe_copy_tolerations(
    tolerations: list[kubernetes.client.V1Toleration],
) -> list[kubernetes.client.V1Toleration]:
    """
    Safely copy a list of V1Toleration objects without mutating the originals.

    Explicitly reconstructs V1Toleration objects instead of using deepcopy() to avoid
    serialization errors with K8s client objects that contain threading primitives
    and non-copyable elements like RLock objects.

    Args:
        tolerations: List of V1Toleration objects to copy

    Returns:
        New list containing copied V1Toleration objects with identical field values"""
    return [
        kubernetes.client.V1Toleration(
            effect=toleration.effect,
            key=toleration.key,
            value=toleration.value,
            operator=toleration.operator,
            toleration_seconds=toleration.toleration_seconds,
        )
        for toleration in tolerations
    ]


def _safe_copy_affinity(
    affinity: kubernetes.client.V1Affinity,
) -> kubernetes.client.V1Affinity:
    """
    Safely create a deep copy of a V1Affinity object.

    Uses K8s API client serialization/deserialization instead of deepcopy() to avoid
    errors with threading primitives and complex internal structures in K8s objects.
    Serializes to dict then deserializes back to a clean V1Affinity object.

    Args:
        affinity: V1Affinity object to copy, or None

    Returns:
        New V1Affinity object with identical field values, or None if input was None
    """
    if not affinity:
        return None
    api_client = kubernetes.client.ApiClient()
    # Convert to dict then back to object properly
    affinity_dict = api_client.sanitize_for_serialization(affinity)
    return api_client._ApiClient__deserialize(affinity_dict, "V1Affinity")


def _get_mode_handler(mode: str):
    return {
        mlrun.common.schemas.PreemptionModes.prevent: _handle_prevent_mode,
        mlrun.common.schemas.PreemptionModes.constrain: _handle_constrain_mode,
        mlrun.common.schemas.PreemptionModes.allow: _handle_allow_mode,
    }.get(mode)


def _handle_prevent_mode(
    node_selector: dict[str, str],
    tolerations: list[kubernetes.client.V1Toleration],
    affinity: typing.Optional[kubernetes.client.V1Affinity],
    preemptible_tolerations: list[kubernetes.client.V1Toleration],
) -> tuple[
    dict[str, str],
    list[kubernetes.client.V1Toleration],
    typing.Optional[kubernetes.client.V1Affinity],
]:
    # Ensure no preemptible node tolerations
    tolerations = [t for t in tolerations if t not in preemptible_tolerations]

    # Purge affinity preemption-related configuration
    affinity = _prune_affinity_node_selector_requirement(
        generate_preemptible_node_selector_requirements(
            mlrun.common.schemas.NodeSelectorOperator.node_selector_op_in.value
        ),
        affinity=affinity,
    )

    # Remove preemptible nodes constraint
    node_selector = _prune_node_selector(
        mlconfig.get_preemptible_node_selector(),
        enriched_node_selector=node_selector,
    )

    # Use anti-affinity only if no tolerations configured
    if not preemptible_tolerations:
        affinity = _override_required_during_scheduling_ignored_during_execution(
            kubernetes.client.V1NodeSelector(
                node_selector_terms=generate_preemptible_nodes_anti_affinity_terms()
            ),
            affinity,
        )

    return node_selector, tolerations, affinity


def _handle_constrain_mode(
    node_selector: dict[str, str],
    tolerations: list[kubernetes.client.V1Toleration],
    affinity: typing.Optional[kubernetes.client.V1Affinity],
    preemptible_tolerations: list[kubernetes.client.V1Toleration],
) -> tuple[
    dict[str, str],
    list[kubernetes.client.V1Toleration],
    typing.Optional[kubernetes.client.V1Affinity],
]:
    tolerations = _merge_tolerations(tolerations, preemptible_tolerations)

    affinity = _override_required_during_scheduling_ignored_during_execution(
        kubernetes.client.V1NodeSelector(
            node_selector_terms=generate_preemptible_nodes_affinity_terms()
        ),
        affinity=affinity,
    )

    return node_selector, tolerations, affinity


def _handle_allow_mode(
    node_selector: dict[str, str],
    tolerations: list[kubernetes.client.V1Toleration],
    affinity: typing.Optional[kubernetes.client.V1Affinity],
    preemptible_tolerations: list[kubernetes.client.V1Toleration],
) -> tuple[
    dict[str, str],
    list[kubernetes.client.V1Toleration],
    typing.Optional[kubernetes.client.V1Affinity],
]:
    tolerations = _merge_tolerations(tolerations, preemptible_tolerations)
    return node_selector, tolerations, affinity


def _merge_tolerations(
    existing: list[kubernetes.client.V1Toleration],
    to_add: list[kubernetes.client.V1Toleration],
) -> list[kubernetes.client.V1Toleration]:
    for toleration in to_add:
        if toleration not in existing:
            existing.append(toleration)
    return existing


def _prune_node_selector(
    node_selector: dict[str, str],
    enriched_node_selector: dict[str, str],
):
    """
    Prunes given node_selector key from function spec if their key and value are matching
    :param node_selector: node selectors to prune
    """
    # both needs to exists to prune required node_selector from the spec node selector
    if not node_selector or not enriched_node_selector:
        return

    mlrun.utils.logger.debug("Pruning node selectors", node_selector=node_selector)
    return {
        key: value
        for key, value in enriched_node_selector.items()
        if node_selector.get(key) != value
    }


def _prune_affinity_node_selector_requirement(
    node_selector_requirements: list[kubernetes.client.V1NodeSelectorRequirement],
    affinity: typing.Optional[kubernetes.client.V1Affinity],
):
    """
    Prunes given node selector requirements from affinity.
    We are only editing required_during_scheduling_ignored_during_execution because the scheduler can't schedule
    the pod unless the rule is met.
    :param node_selector_requirements:
    :return:
    """
    # both needs to exist to prune required affinity from spec affinity
    if not affinity or not node_selector_requirements:
        return
    if affinity.node_affinity:
        node_affinity: kubernetes.client.V1NodeAffinity = affinity.node_affinity

        new_required_during_scheduling_ignored_during_execution = None
        if node_affinity.required_during_scheduling_ignored_during_execution:
            node_selector: kubernetes.client.V1NodeSelector = (
                node_affinity.required_during_scheduling_ignored_during_execution
            )
            new_node_selector_terms = (
                _prune_node_selector_requirements_from_node_selector_terms(
                    node_selector_terms=node_selector.node_selector_terms,
                    requirements_to_prune=node_selector_requirements,
                )
            )
            # check whether there are node selector terms to add to the new list of required terms
            if new_node_selector_terms:
                new_required_during_scheduling_ignored_during_execution = (
                    kubernetes.client.V1NodeSelector(
                        node_selector_terms=new_node_selector_terms
                    )
                )
        # if both preferred and new required are empty, clean node_affinity
        if (
            not node_affinity.preferred_during_scheduling_ignored_during_execution
            and not new_required_during_scheduling_ignored_during_execution
        ):
            affinity.node_affinity = None
            return

        _initialize_affinity(affinity=affinity)
        _initialize_node_affinity(affinity=affinity)

        affinity.node_affinity.required_during_scheduling_ignored_during_execution = (
            new_required_during_scheduling_ignored_during_execution
        )
        return affinity


def _prune_node_selector_requirements_from_node_selector_terms(
    node_selector_terms: list[kubernetes.client.V1NodeSelectorTerm],
    requirements_to_prune: list[kubernetes.client.V1NodeSelectorRequirement],
) -> list[kubernetes.client.V1NodeSelectorTerm]:
    """
    Removes matching node selector requirements from the given list of node selector terms.

    Each term may contain multiple match expressions. This function iterates over each expression,
    and removes any that exactly match one of the requirements provided.

    :param node_selector_terms: List of V1NodeSelectorTerm objects to be processed.
    :param requirements_to_prune: List of V1NodeSelectorRequirement objects to remove.
    :return: A new list of V1NodeSelectorTerm objects with the specified requirements pruned.
    """
    pruned_terms = []

    for term in node_selector_terms:
        remaining_requirements = [
            expr
            for expr in term.match_expressions or []
            if expr not in requirements_to_prune
        ]

        # Only add term if there are remaining match expressions or match fields
        if remaining_requirements or term.match_fields:
            pruned_terms.append(
                kubernetes.client.V1NodeSelectorTerm(
                    match_expressions=remaining_requirements,
                    match_fields=term.match_fields,
                )
            )

    return pruned_terms


def _override_required_during_scheduling_ignored_during_execution(
    node_selector: kubernetes.client.V1NodeSelector,
    affinity: typing.Optional[kubernetes.client.V1Affinity],
):
    affinity = _initialize_affinity(affinity)
    affinity = _initialize_node_affinity(affinity)
    affinity.node_affinity.required_during_scheduling_ignored_during_execution = (
        node_selector
    )
    return affinity


def _initialize_affinity(
    affinity: typing.Optional[kubernetes.client.V1Affinity],
) -> kubernetes.client.V1Affinity:
    return affinity or kubernetes.client.V1Affinity()


def _initialize_node_affinity(
    affinity: typing.Optional[kubernetes.client.V1Affinity],
) -> kubernetes.client.V1Affinity:
    affinity = affinity or kubernetes.client.V1Affinity()
    affinity.node_affinity = (
        affinity.node_affinity or kubernetes.client.V1NodeAffinity()
    )
    return affinity


def _prune_empty_affinity(
    affinity: typing.Optional[kubernetes.client.V1Affinity],
) -> typing.Optional[kubernetes.client.V1Affinity]:
    """
    Return None if the given affinity object has no meaningful constraints.

    Keeps the affinity object only if it contains:
    - Any pod affinity or pod anti-affinity
    - Preferred node affinity
    - Required node affinity with at least one match expression or match field
    """
    if not affinity:
        return None

    node_affinity = affinity.node_affinity
    pod_affinity = affinity.pod_affinity
    pod_anti_affinity = affinity.pod_anti_affinity

    # If any pod affinity exists, keep the object
    if pod_affinity or pod_anti_affinity:
        return affinity

    # If node affinity exists, check if it has any meaningful content
    if node_affinity:
        required = node_affinity.required_during_scheduling_ignored_during_execution
        preferred = node_affinity.preferred_during_scheduling_ignored_during_execution

        if preferred:
            return affinity

        if required and required.node_selector_terms:
            for term in required.node_selector_terms:
                if term.match_expressions or term.match_fields:
                    return affinity  # at least one term has meaningful constraints

    # At this point, none of the affinity sections contain meaningful constraints,
    # so the affinity object is effectively empty and can be safely discarded.
    return None



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import ast
import copy
import os
import uuid
from typing import Any, Callable, Optional, Union

import mlrun.common.constants
import mlrun.common.runtimes.constants
import mlrun.common.schemas
import mlrun.config
import mlrun.errors
import mlrun.lists
import mlrun.model
import mlrun.runtimes
import mlrun.utils.regex
import mlrun_pipelines.common.ops
from mlrun.utils import logger

run_modes = ["pass"]


class BaseLauncher(abc.ABC):
    """
    Abstract class for managing and running functions in different contexts
    This class is designed to encapsulate the logic of running a function in different contexts
    i.e. running a function locally, remotely or in a server
    Each context will have its own implementation of the abstract methods while the common logic resides in this class
    """

    def __init__(self, **kwargs):
        pass

    @abc.abstractmethod
    def launch(
        self,
        runtime: "mlrun.runtimes.BaseRuntime",
        task: Optional[
            Union["mlrun.run.RunTemplate", "mlrun.run.RunObject", dict]
        ] = None,
        handler: Optional[Union[str, Callable]] = None,
        name: Optional[str] = "",
        project: Optional[str] = "",
        params: Optional[dict] = None,
        inputs: Optional[dict[str, str]] = None,
        out_path: Optional[str] = "",
        workdir: Optional[str] = "",
        artifact_path: Optional[str] = "",
        output_path: Optional[str] = "",
        watch: Optional[bool] = True,
        schedule: Optional[
            Union[str, mlrun.common.schemas.schedule.ScheduleCronTrigger]
        ] = None,
        hyperparams: Optional[dict[str, list]] = None,
        hyper_param_options: Optional[mlrun.model.HyperParamOptions] = None,
        verbose: Optional[bool] = None,
        scrape_metrics: Optional[bool] = None,
        local_code_path: Optional[str] = None,
        auto_build: Optional[bool] = None,
        param_file_secrets: Optional[dict[str, str]] = None,
        notifications: Optional[list[mlrun.model.Notification]] = None,
        returns: Optional[list[Union[str, dict[str, str]]]] = None,
        state_thresholds: Optional[dict[str, int]] = None,
        retry: Optional[Union[mlrun.model.Retry, dict]] = None,
    ) -> "mlrun.run.RunObject":
        """run the function from the server/client[local/remote]"""
        pass

    @abc.abstractmethod
    def enrich_runtime(
        self,
        runtime: "mlrun.runtimes.base.BaseRuntime",
        project_name: Optional[str] = "",
        full: bool = True,
        client_version: str = "",
    ):
        pass

    @staticmethod
    @abc.abstractmethod
    def _store_function(
        runtime: "mlrun.runtimes.BaseRuntime", run: "mlrun.run.RunObject"
    ):
        pass

    def save_function(
        self,
        runtime: "mlrun.runtimes.BaseRuntime",
        tag: str = "",
        versioned: bool = False,
        refresh: bool = False,
    ) -> str:
        """
        store the function to the db
        :param runtime:     runtime object
        :param tag:         function tag to store
        :param versioned:   whether we want to version this function object so that it will queryable by its hash key
        :param refresh:     refresh function metadata

        :return:            function uri
        """
        db = runtime._get_db()
        if not db:
            raise mlrun.errors.MLRunPreconditionFailedError(
                "Database connection is not configured"
            )

        if refresh:
            self._refresh_function_metadata(runtime)

        tag = tag or runtime.metadata.tag

        obj = runtime.to_dict()
        logger.debug("Saving function", runtime_name=runtime.metadata.name, tag=tag)
        hash_key = db.store_function(
            obj, runtime.metadata.name, runtime.metadata.project, tag, versioned
        )
        hash_key = hash_key if versioned else None
        return "db://" + runtime._function_uri(hash_key=hash_key, tag=tag)

    @staticmethod
    def prepare_image_for_deploy(runtime: "mlrun.runtimes.BaseRuntime"):
        """Check if the runtime requires to build the image and updates the spec accordingly"""
        pass

    def _validate_run(
        self,
        runtime: "mlrun.runtimes.BaseRuntime",
        run: "mlrun.run.RunObject",
    ):
        mlrun.utils.helpers.verify_dict_items_type(
            "Inputs", run.spec.inputs, [str], [str]
        )

        if runtime.spec.mode and runtime.spec.mode not in run_modes:
            raise ValueError(f'run mode can only be {",".join(run_modes)}')

        self._validate_run_params(run.spec.parameters)
        self._validate_output_path(runtime, run)

        for image in [
            runtime.spec.image,
            getattr(runtime.spec.build, "base_image", None),
        ]:
            mlrun.utils.helpers.warn_on_deprecated_image(image)

        # Raise an error if retry is configured for a runtime that doesn't support retries.
        # For local runs, we intentionally skip this validation and allow the run to proceed, since they are typically
        # used for debugging purposes, and in such cases we avoid blocking their execution.
        if (
            not mlrun.runtimes.RuntimeKinds.is_local_runtime(runtime.kind)
            and run.spec.retry.count
            and runtime.kind not in mlrun.runtimes.RuntimeKinds.retriable_runtimes()
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Retry is not supported for {runtime.kind} runtime, supported runtimes are: "
                f"{mlrun.runtimes.RuntimeKinds.retriable_runtimes()}"
            )

    @staticmethod
    def _validate_output_path(
        runtime: "mlrun.runtimes.BaseRuntime",
        run: "mlrun.run.RunObject",
    ):
        if not run.spec.output_path or "://" not in run.spec.output_path:
            message = ""
            if not os.path.isabs(run.spec.output_path):
                message = (
                    "Artifact/output path is not defined or is local and relative,"
                    " artifacts will not be visible in the UI"
                )
                if mlrun.runtimes.RuntimeKinds.requires_absolute_artifacts_path(
                    runtime.kind
                ):
                    raise mlrun.errors.MLRunPreconditionFailedError(
                        "Artifact path (`artifact_path`) must be absolute for remote tasks"
                    )
            elif (
                hasattr(runtime.spec, "volume_mounts")
                and not runtime.spec.volume_mounts
            ):
                message = (
                    "Artifact output path is local while no volume mount is specified. "
                    "Artifacts would not be visible via UI."
                )
            if message:
                logger.warning(message, output_path=run.spec.output_path)

    def _validate_run_params(self, parameters: dict[str, Any]):
        for param_name, param_value in parameters.items():
            if isinstance(param_value, dict):
                # if the parameter is a dict, we might have some nested parameters,
                # in this case we need to verify them as well recursively
                self._validate_run_params(param_value)
            self._validate_run_single_param(
                param_name=param_name, param_value=param_value
            )

    @classmethod
    def _validate_run_single_param(cls, param_name: str, param_value: int):
        # verify that integer parameters don't exceed a int64
        if isinstance(param_value, int) and abs(param_value) >= 2**63:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Parameter {param_name} value {param_value} exceeds int64"
            )

    @staticmethod
    def _create_run_object(task):
        if not task:
            # if task passed generate default RunObject
            return mlrun.run.RunObject.from_dict(task)

        # deepcopy user's task, so we don't modify / enrich the user's object
        task = copy.deepcopy(task)

        if isinstance(task, str):
            task = ast.literal_eval(task)

        valid_task_types = (dict, mlrun.run.RunTemplate, mlrun.run.RunObject)
        if isinstance(task, mlrun.run.RunObject):
            # if task is already a RunObject, we can return it as is
            return task
        if isinstance(task, mlrun.run.RunTemplate):
            return mlrun.run.RunObject.from_template(task)
        elif isinstance(task, dict):
            return mlrun.run.RunObject.from_dict(task)

        raise mlrun.errors.MLRunInvalidArgumentError(
            f"Task is not a valid object, type={type(task)}, expected types={valid_task_types}"
        )

    @staticmethod
    def _enrich_run(
        runtime,
        run,
        handler=None,
        project_name=None,
        name=None,
        params=None,
        inputs=None,
        returns=None,
        hyperparams=None,
        hyper_param_options=None,
        verbose=None,
        scrape_metrics=None,
        output_path=None,
        workdir=None,
        notifications: Optional[list[mlrun.model.Notification]] = None,
        state_thresholds: Optional[dict[str, int]] = None,
        retry: Optional[Union[mlrun.model.Retry, dict]] = None,
    ):
        run.spec.handler = (
            handler or run.spec.handler or runtime.spec.default_handler or ""
        )
        # callable handlers are valid for handler and dask runtimes,
        # for other runtimes we need to convert the handler to a string
        if run.spec.handler and runtime.kind not in ["handler", "dask"]:
            run.spec.handler = run.spec.handler_name

        def_name = runtime.metadata.name
        if run.spec.handler_name:
            short_name = run.spec.handler_name
            for separator in ["#", "::", "."]:
                # drop paths, module or class name from short name
                if separator in short_name:
                    short_name = short_name.split(separator)[-1]
            def_name += "-" + short_name

        run.metadata.name = mlrun.utils.normalize_name(
            name=name or run.metadata.name or def_name,
        )
        mlrun.utils.verify_field_regex(
            "run.metadata.name", run.metadata.name, mlrun.utils.regex.run_name
        )
        run.metadata.project = (
            project_name
            or run.metadata.project
            or runtime.metadata.project
            or mlrun.mlconf.active_project
        )
        run.spec.parameters = params or run.spec.parameters
        run.spec.inputs = inputs or run.spec.inputs
        run.spec.returns = returns or run.spec.returns
        run.spec.hyperparams = hyperparams or run.spec.hyperparams
        run.spec.hyper_param_options = (
            hyper_param_options or run.spec.hyper_param_options
        )
        run.spec.verbose = verbose or run.spec.verbose
        if scrape_metrics is None:
            if run.spec.scrape_metrics is None:
                scrape_metrics = mlrun.mlconf.scrape_metrics
            else:
                scrape_metrics = run.spec.scrape_metrics
        run.spec.scrape_metrics = scrape_metrics
        run.spec.input_path = workdir or run.spec.input_path or runtime.spec.workdir
        if runtime.spec.allow_empty_resources:
            run.spec.allow_empty_resources = runtime.spec.allow_empty_resources

        spec = run.spec
        if spec.secret_sources:
            runtime._secrets = mlrun.secrets.SecretsStore.from_list(spec.secret_sources)

        # update run metadata (uid, labels) and store in DB
        meta = run.metadata
        meta.uid = meta.uid or uuid.uuid4().hex

        run.spec.output_path = output_path or run.spec.output_path

        if not run.spec.output_path:
            if run.metadata.project:
                if (
                    mlrun.pipeline_context.project
                    and run.metadata.project
                    == mlrun.pipeline_context.project.metadata.name
                ):
                    run.spec.output_path = (
                        mlrun.pipeline_context.project.spec.artifact_path
                        or mlrun.pipeline_context.workflow_artifact_path
                    )

                # get_db might be None when no rundb is set on runtime
                if not run.spec.output_path and runtime._get_db():
                    try:
                        # not passing or loading the DB before the enrichment on purpose, because we want to enrich the
                        # spec first as get_db() depends on it
                        project = runtime._get_db().get_project(run.metadata.project)
                        # this is mainly for tests, so we won't need to mock get_project for so many tests
                        # in normal use cases if no project is found we will get an error
                        if project:
                            run.spec.output_path = project.spec.artifact_path
                    except mlrun.errors.MLRunNotFoundError:
                        logger.warning(
                            f"Project {project_name} is not saved in DB yet, "
                            f"enriching output path with default artifact path: {mlrun.mlconf.artifact_path}"
                        )

            if not run.spec.output_path:
                run.spec.output_path = mlrun.mlconf.artifact_path

        if run.spec.output_path:
            run.spec.output_path = mlrun.utils.helpers.template_artifact_path(
                run.spec.output_path, run.metadata.project, meta.uid
            )

        notifications = notifications or run.spec.notifications or []
        mlrun.model.Notification.validate_notification_uniqueness(notifications)
        for notification in notifications:
            notification.validate_notification()

        run.spec.notifications = notifications

        state_thresholds = (
            state_thresholds
            or run.spec.state_thresholds
            or getattr(runtime.spec, "state_thresholds", {})
            or {}
        )
        state_thresholds = (
            mlrun.mlconf.function.spec.state_thresholds.default.to_dict()
            | state_thresholds
        )
        run.spec.state_thresholds = state_thresholds or run.spec.state_thresholds
        run.spec.retry = retry or run.spec.retry
        return run

    @staticmethod
    def _run_has_valid_notifications(runobj) -> bool:
        if not runobj.spec.notifications:
            logger.debug(
                "No notifications to push for run", run_uid=runobj.metadata.uid
            )
            return False

        # TODO: add support for other notifications per run iteration
        if runobj.metadata.iteration and runobj.metadata.iteration > 0:
            logger.debug(
                "Notifications per iteration are not supported, skipping",
                run_uid=runobj.metadata.uid,
            )
            return False

        return True

    def _wrap_run_result(
        self,
        runtime: "mlrun.runtimes.BaseRuntime",
        result: dict,
        run: "mlrun.run.RunObject",
        schedule: Optional[mlrun.common.schemas.ScheduleCronTrigger] = None,
        err: Optional[Exception] = None,
    ):
        # if the purpose was to schedule (and not to run) nothing to wrap
        if schedule:
            return

        if result and runtime.kfp and err is None:
            mlrun_pipelines.common.ops.write_kfpmeta(result)

        self._log_track_results(runtime.is_child, result, run)

        if result:
            run = mlrun.run.RunObject.from_dict(result)
            logger.info(
                "Run execution finished",
                status=run.status.state,
                name=run.metadata.name,
            )
            if (
                run.status.state
                in mlrun.common.runtimes.constants.RunStates.error_states()
            ):
                if runtime._is_remote and not runtime.is_child:
                    logger.error(
                        "Run did not finish successfully",
                        state=run.status.state,
                        status=run.status.to_dict(),
                    )

                error = run.error
                if (
                    run.status.state
                    == mlrun.common.runtimes.constants.RunStates.pending_retry
                ):
                    error = f"Run is pending retry, error: {run.error}"
                raise mlrun.runtimes.utils.RunError(error)
            return run

        return None

    @staticmethod
    def _refresh_function_metadata(runtime: "mlrun.runtimes.BaseRuntime"):
        pass

    @staticmethod
    def _log_track_results(
        runtime: "mlrun.runtimes.BaseRuntime", result: dict, run: "mlrun.run.RunObject"
    ):
        pass



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import json
from datetime import datetime
from typing import Any, Literal, NamedTuple, Optional, TypeVar
from uuid import UUID

from pydantic import validator  # use `validator` if you’re still on Pydantic v1
from pydantic.v1 import BaseModel, Field, constr

# TODO: remove the unused import below after `mlrun.datastore` and `mlrun.utils` usage is removed.
# At the moment `make lint` fails if this is removed.
from ..object import ObjectKind, ObjectMetadata, ObjectSpec, ObjectStatus
from . import ModelEndpointSchema
from .constants import (
    FQN_REGEX,
    MODEL_ENDPOINT_ID_PATTERN,
    PROJECT_PATTERN,
    EndpointMode,
    EndpointType,
    ModelEndpointMonitoringMetricType,
    ModelMonitoringMode,
    ResultKindApp,
    ResultStatusApp,
)

Model = TypeVar("Model", bound=BaseModel)


class Histogram(BaseModel):
    buckets: list[float]
    counts: list[int]


class FeatureValues(BaseModel):
    min: float
    mean: float
    max: float
    histogram: Histogram

    @classmethod
    def from_dict(cls, stats: Optional[dict]):
        if stats:
            return FeatureValues(
                min=stats["min"],
                mean=stats["mean"],
                max=stats["max"],
                histogram=Histogram(buckets=stats["hist"][1], counts=stats["hist"][0]),
            )
        else:
            return None


class Features(BaseModel):
    name: str
    weight: float
    expected: Optional[FeatureValues]
    actual: Optional[FeatureValues]

    @classmethod
    def new(
        cls,
        feature_name: str,
        feature_stats: Optional[dict],
        current_stats: Optional[dict],
    ):
        return cls(
            name=feature_name,
            weight=-1.0,
            expected=FeatureValues.from_dict(feature_stats),
            actual=FeatureValues.from_dict(current_stats),
        )


class ModelEndpointParser(abc.ABC, BaseModel):
    @classmethod
    def json_parse_values(cls) -> list[str]:
        return []

    @classmethod
    def from_flat_dict(
        cls,
        endpoint_dict: dict,
        json_parse_values: Optional[list] = None,
        validate: bool = True,
    ) -> "ModelEndpointParser":
        """Create a `ModelEndpointParser` object from an endpoint dictionary

        :param endpoint_dict:     Model endpoint dictionary.
        :param json_parse_values: List of dictionary keys with a JSON string value that will be parsed into a
                                  dictionary using json.loads().
        :param validate:          Whether to validate the flattened dictionary.
                                  Skip validation to optimize performance when it is safe to do so.
        """
        if json_parse_values is None:
            json_parse_values = cls.json_parse_values()

        return _mapping_attributes(
            model_class=cls,
            flattened_dictionary=endpoint_dict,
            json_parse_values=json_parse_values,
            validate=validate,
        )


class ModelEndpointMetadata(ObjectMetadata, ModelEndpointParser):
    project: constr(regex=PROJECT_PATTERN)
    endpoint_type: EndpointType = EndpointType.NODE_EP
    uid: Optional[constr(regex=MODEL_ENDPOINT_ID_PATTERN)]
    mode: Optional[EndpointMode] = None

    @classmethod
    def mutable_fields(cls):
        return ["labels"]

    @validator("uid", pre=True)
    def _uid_to_str(cls, v):  # noqa: N805
        if isinstance(v, UUID):
            return str(v)
        return v

    @validator("mode", pre=True, always=True)
    def _set_mode_based_on_endpoint_type(cls, v, values):  # noqa: N805
        if v is None:
            if values.get("endpoint_type") == EndpointType.BATCH_EP:
                return EndpointMode.BATCH_LEGACY
            else:
                return EndpointMode.REAL_TIME
        return v


class ModelEndpointSpec(ObjectSpec, ModelEndpointParser):
    model_class: Optional[str] = ""
    function_name: Optional[str] = ""
    function_tag: Optional[str] = ""
    model_path: Optional[str] = ""
    model_name: Optional[str] = ""
    model_tags: Optional[list[str]] = []
    _model_id: Optional[int] = ""
    feature_names: Optional[list[str]] = []
    label_names: Optional[list[str]] = []
    feature_stats: Optional[dict] = {}
    function_uri: Optional[str] = ""  # <project_name>/<function_hash>
    model_uri: Optional[str] = ""
    children: Optional[list[str]] = []
    children_uids: Optional[list[str]] = []
    monitoring_feature_set_uri: Optional[str] = ""

    @classmethod
    def mutable_fields(cls):
        return [
            "model_path",
            "model_class",
            "feature_names",
            "label_names",
            "children",
            "children_uids",
        ]


class ModelEndpointStatus(ObjectStatus, ModelEndpointParser):
    state: Optional[str] = "unknown"  # will be updated according to the function state
    first_request: Optional[datetime] = None
    monitoring_mode: Optional[ModelMonitoringMode] = ModelMonitoringMode.disabled
    sampling_percentage: Optional[float] = 100

    # operative
    last_request: Optional[datetime] = None
    result_status: Optional[int] = -1
    avg_latency: Optional[float] = None
    error_count: Optional[int] = 0
    current_stats: Optional[dict] = {}
    current_stats_timestamp: Optional[datetime] = None
    drift_measures: Optional[dict] = {}
    drift_measures_timestamp: Optional[datetime] = None

    @classmethod
    def mutable_fields(cls):
        return [
            "monitoring_mode",
            "first_request",
            "last_request",
            "sampling_percentage",
        ]


class ModelEndpoint(BaseModel):
    kind: ObjectKind = Field(ObjectKind.model_endpoint, const=True)
    metadata: ModelEndpointMetadata
    spec: ModelEndpointSpec
    status: ModelEndpointStatus

    @classmethod
    def mutable_fields(cls):
        return (
            ModelEndpointMetadata.mutable_fields()
            + ModelEndpointSpec.mutable_fields()
            + ModelEndpointStatus.mutable_fields()
        )

    def flat_dict(self) -> dict[str, Any]:
        """Generate a flattened `ModelEndpoint` dictionary. The flattened dictionary result is important for storing
        the model endpoint object in the database.

        :return: Flattened `ModelEndpoint` dictionary.
        """
        # Convert the ModelEndpoint object into a dictionary using BaseModel dict() function
        # In addition, remove the BaseModel kind as it is not required by the DB schema

        model_endpoint_dictionary = self.dict(exclude={"kind"})
        exclude = {
            "tag",
            ModelEndpointSchema.FEATURE_STATS,
            ModelEndpointSchema.CURRENT_STATS,
            ModelEndpointSchema.DRIFT_MEASURES,
            ModelEndpointSchema.FUNCTION_URI,
        }
        # Initialize a flattened dictionary that will be filled with the model endpoint dictionary attributes
        flatten_dict = {}
        for k_object in model_endpoint_dictionary:
            for key in model_endpoint_dictionary[k_object]:
                if key not in exclude:
                    # Extract the value of the current field
                    flatten_dict[key] = model_endpoint_dictionary[k_object][key]

        return flatten_dict

    @classmethod
    def from_flat_dict(
        cls, endpoint_dict: dict, validate: bool = True
    ) -> "ModelEndpoint":
        """Create a `ModelEndpoint` object from an endpoint flattened dictionary. Because the provided dictionary
        is flattened, we pass it as is to the subclasses without splitting the keys into spec, metadata, and status.

        :param endpoint_dict:     Model endpoint dictionary.
        :param validate:          Whether to validate the flattened dictionary.
                                  Skip validation to optimize performance when it is safe to do so.
        """

        return cls(
            metadata=ModelEndpointMetadata.from_flat_dict(
                endpoint_dict=endpoint_dict, validate=validate
            ),
            spec=ModelEndpointSpec.from_flat_dict(
                endpoint_dict=endpoint_dict, validate=validate
            ),
            status=ModelEndpointStatus.from_flat_dict(
                endpoint_dict=endpoint_dict, validate=validate
            ),
        )

    def get(self, field, default=None):
        return (
            getattr(self.metadata, field, None)
            or getattr(self.spec, field, None)
            or getattr(self.status, field, None)
            or default
        )


class ModelEndpointList(BaseModel):
    endpoints: list[ModelEndpoint]


class ModelEndpointMonitoringMetric(BaseModel):
    project: str
    app: str
    type: ModelEndpointMonitoringMetricType
    name: str
    full_name: Optional[str] = None
    kind: Optional[ResultKindApp] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.full_name = compose_full_name(
            project=self.project, app=self.app, name=self.name, type=self.type
        )


def compose_full_name(
    *,
    project: str,
    app: str,
    name: str,
    type: ModelEndpointMonitoringMetricType = ModelEndpointMonitoringMetricType.RESULT,
) -> str:
    return ".".join([project, app, type, name])


def _parse_metric_fqn_to_monitoring_metric(fqn: str) -> ModelEndpointMonitoringMetric:
    match = FQN_REGEX.fullmatch(fqn)
    if match is None:
        raise ValueError("The fully qualified name is not in the expected format")
    return ModelEndpointMonitoringMetric.parse_obj(
        match.groupdict() | {"full_name": fqn}
    )


class _MetricPoint(NamedTuple):
    timestamp: datetime
    value: float


class _ResultPoint(NamedTuple):
    timestamp: datetime
    value: float
    status: ResultStatusApp
    extra_data: Optional[str] = ""


class _ModelEndpointMonitoringMetricValuesBase(BaseModel):
    full_name: str
    type: ModelEndpointMonitoringMetricType
    data: bool


class ModelEndpointMonitoringMetricValues(_ModelEndpointMonitoringMetricValuesBase):
    type: ModelEndpointMonitoringMetricType = ModelEndpointMonitoringMetricType.METRIC
    values: list[_MetricPoint]
    data: bool = True


class ModelEndpointMonitoringResultValues(_ModelEndpointMonitoringMetricValuesBase):
    type: ModelEndpointMonitoringMetricType = ModelEndpointMonitoringMetricType.RESULT
    result_kind: ResultKindApp
    values: list[_ResultPoint]
    data: bool = True


class ModelEndpointMonitoringMetricNoData(_ModelEndpointMonitoringMetricValuesBase):
    full_name: str
    type: ModelEndpointMonitoringMetricType
    data: bool = False


class ApplicationBaseRecord(BaseModel):
    type: Literal["metric", "result"]
    value: float
    time: Optional[datetime] = None


class ApplicationResultRecord(ApplicationBaseRecord):
    kind: ResultKindApp
    status: ResultStatusApp
    result_name: str
    type: Literal["result"] = "result"


class ApplicationMetricRecord(ApplicationBaseRecord):
    metric_name: str
    type: Literal["metric"] = "metric"


class _DriftBin(NamedTuple):
    timestamp: datetime
    count_suspected: int
    count_detected: int


class ModelEndpointDriftValues(BaseModel):
    values: list[_DriftBin]


def _mapping_attributes(
    model_class: type[Model],
    flattened_dictionary: dict,
    json_parse_values: list,
    validate: bool = True,
) -> Model:
    """Generate a `BaseModel` object with the provided dictionary attributes.

    :param model_class:          `BaseModel` class (e.g. `ModelEndpointMetadata`).
    :param flattened_dictionary: Flattened dictionary that contains the model endpoint attributes.
    :param json_parse_values:    List of dictionary keys with a JSON string value that will be parsed into a
                                 dictionary using json.loads().
    :param validate:             Whether to validate the flattened dictionary.
                                 Skip validation to optimize performance when it is safe to do so.
    """
    # Get the fields of the provided base model object. These fields will be used to filter to relevant keys
    # from the flattened dictionary.
    wanted_keys = model_class.__fields__.keys()

    # Generate a filtered flattened dictionary that will be parsed into the BaseModel object
    dict_to_parse = {}
    for field_key in wanted_keys:
        if field_key in flattened_dictionary:
            if field_key in json_parse_values:
                # Parse the JSON value into a valid dictionary
                dict_to_parse[field_key] = _json_loads_if_not_none(
                    flattened_dictionary[field_key]
                )
            elif flattened_dictionary[field_key] != "null":
                dict_to_parse[field_key] = flattened_dictionary[field_key]
            else:
                dict_to_parse[field_key] = None

    if validate:
        return model_class.parse_obj(dict_to_parse)

    return model_class.construct(**dict_to_parse)


def _json_loads_if_not_none(field: Any) -> Any:
    return (
        json.loads(field) if field and field != "null" and field is not None else None
    )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import typing
from copy import deepcopy
from typing import Optional

import aiohttp

import mlrun.common.schemas
import mlrun.lists


class TimedHTTPClient:
    def __init__(self, timeout: Optional[float] = 30.0):
        """
        HTTP client wrapper with built-in timeout.

        Args:
            timeout: Request timeout in seconds (default: 30.0)
        """
        self.timeout = aiohttp.ClientTimeout(total=timeout)

    def session(self, **kwargs) -> aiohttp.ClientSession:
        """Create a new ClientSession with the configured timeout and additional parameters."""
        return aiohttp.ClientSession(timeout=self.timeout, **kwargs)


class NotificationBase:
    def __init__(
        self,
        name: typing.Optional[str] = None,
        params: typing.Optional[dict[str, str]] = None,
        default_params: typing.Optional[dict[str, str]] = None,
    ):
        """
        NotificationBase is the base class for all notification types.

        :param name: The name of the notification.
        :param params: The parameters of the notification.
        :param default_params: The default parameters of the notification. Used for server-side enrichment purposes.
        """
        self.name = name
        self.params = params or {}
        self.params = self.enrich_default_params(self.params, default_params)

    @classmethod
    def validate_params(cls, params):
        pass

    @property
    def active(self) -> bool:
        return True

    @property
    def is_async(self) -> bool:
        return asyncio.iscoroutinefunction(self.push)

    def push(
        self,
        message: str,
        severity: typing.Optional[
            typing.Union[mlrun.common.schemas.NotificationSeverity, str]
        ] = mlrun.common.schemas.NotificationSeverity.INFO,
        runs: typing.Optional[typing.Union[mlrun.lists.RunList, list]] = None,
        custom_html: typing.Optional[str] = None,
        alert: typing.Optional[mlrun.common.schemas.AlertConfig] = None,
        event_data: typing.Optional[mlrun.common.schemas.Event] = None,
    ):
        raise NotImplementedError()

    def load_notification(
        self,
        params: dict[str, str],
    ) -> None:
        self.params = params or {}

    @classmethod
    def enrich_default_params(
        cls, params: dict, default_params: typing.Optional[dict] = None
    ) -> dict:
        default_params = default_params or {}
        returned_params = deepcopy(default_params)
        returned_params.update(params)
        return returned_params

    def _get_html(
        self,
        message: str,
        severity: typing.Optional[
            typing.Union[mlrun.common.schemas.NotificationSeverity, str]
        ] = mlrun.common.schemas.NotificationSeverity.INFO,
        runs: typing.Optional[typing.Union[mlrun.lists.RunList, list]] = None,
        custom_html: typing.Optional[typing.Optional[str]] = None,
        alert: typing.Optional[mlrun.common.schemas.AlertConfig] = None,
        event_data: typing.Optional[mlrun.common.schemas.Event] = None,
    ) -> str:
        if custom_html:
            return custom_html

        if alert:
            if not event_data:
                return f"[{severity}] {message}"

            html = f"<h3>[{severity}] {message}</h3>"
            html += f"<br>{alert.name} alert has occurred<br>"
            html += f"<br><h4>Project:</h4>{alert.project}<br>"
            html += f"<br><h4>ID:</h4>{event_data.entity.ids[0]}<br>"
            html += f"<br><h4>Summary:</h4>{mlrun.utils.helpers.format_alert_summary(alert, event_data)}<br>"

            if event_data.value_dict:
                html += "<br><h4>Event data:</h4>"
                for key, value in event_data.value_dict.items():
                    html += f"{key}: {value}<br>"

            overview_type, url = self._get_overview_type_and_url(alert, event_data)
            html += f"<br><h4>Overview:</h4><a href={url}>{overview_type}</a>"
            return html

        if self.name:
            message = f"{self.name}: {message}"

        if not runs:
            return f"[{severity}] {message}"

        if isinstance(runs, list):
            runs = mlrun.lists.RunList(runs)

        html = f"<h2>Run Results</h2><h3>[{severity}] {message}</h3>"
        html += "<br>click the hyper links below to see detailed results<br>"
        html += runs.show(display=False, short=True)
        return html

    def _get_overview_type_and_url(
        self,
        alert: mlrun.common.schemas.AlertConfig,
        event_data: mlrun.common.schemas.Event,
    ) -> (str, str):
        if (
            event_data.entity.kind == mlrun.common.schemas.alert.EventEntityKind.JOB
        ):  # JOB entity
            uid = event_data.value_dict.get("uid")
            name = event_data.entity.ids[0]
            url = mlrun.utils.helpers.get_run_url(
                alert.project,
                uid=uid,
                name=name,
            )
            overview_type = "Job overview"
        else:  # MODEL entity
            model_name = event_data.value_dict.get("model")
            model_endpoint_id = event_data.value_dict.get("model_endpoint_id")
            url = mlrun.utils.helpers.get_model_endpoint_url(
                alert.project, model_name, model_endpoint_id
            )
            overview_type = "Model endpoint"

        return overview_type, url



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import typing
from datetime import datetime

import pandas as pd

import mlrun
from mlrun.datastore.targets import CSVTarget, ParquetTarget
from mlrun.feature_store.feature_set import FeatureSet
from mlrun.feature_store.feature_vector_utils import JoinGraph, OfflineVectorResponse

from ...utils import logger, str_to_timestamp


class BaseMerger(abc.ABC):
    """abstract feature merger class"""

    # In order to be an online merger, the merger should implement `init_online_vector_service` function.
    support_online = False

    # In order to be an offline merger, the merger should implement
    # `_order_by`, `_filter`, `_drop_columns_from_result`, `_rename_columns_and_select`, `_get_engine_df` functions.
    support_offline = False
    engine = None

    def __init__(self, vector, **engine_args):
        self._relation = dict()
        self._join_type = "inner"
        self._default_join_type = "default_join"
        self.vector = vector

        self._result_df = None
        self._drop_columns = []
        self._index_columns = []
        self._drop_indexes = True
        self._target = None
        self._alias = dict()
        self._origin_alias = dict()
        self._entity_rows_node_name = "__mlrun__$entity_rows$"

    def _append_drop_column(self, key):
        if key and key not in self._drop_columns:
            self._drop_columns.append(key)

    def _append_index(self, key):
        if key:
            if key not in self._index_columns:
                self._index_columns.append(key)
            if self._drop_indexes:
                self._append_drop_column(key)

    def _update_alias(
        self,
        key: typing.Optional[str] = None,
        val: typing.Optional[str] = None,
        dictionary: typing.Optional[dict] = None,
    ):
        if dictionary is not None:
            # adding dictionary to alias
            self._alias.update(dictionary)
        elif val in self._alias.values():
            # changing alias key
            old_key = [key for key, v in self._alias.items() if v == val][0]
            self._alias[key] = self._alias.pop(old_key)
        else:
            self._alias[key] = val

    def start(
        self,
        entity_rows=None,
        entity_timestamp_column=None,
        target=None,
        drop_columns=None,
        start_time=None,
        end_time=None,
        timestamp_for_filtering=None,
        with_indexes=None,
        update_stats=None,
        query=None,
        order_by=None,
        additional_filters=None,
    ):
        self._target = target

        # calculate the index columns and columns we need to drop
        self._drop_columns = drop_columns or self._drop_columns
        if self.vector.spec.with_indexes or with_indexes:
            self._drop_indexes = False

        # retrieve the feature set objects/fields needed for the vector
        feature_set_objects, feature_set_fields = self.vector.parse_features(
            update_stats=update_stats
        )
        if len(feature_set_fields) == 0:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "No features in vector. Make sure to infer the schema on all the feature sets first"
            )

        if update_stats:
            # update the feature vector objects with refreshed stats
            self.vector.save()

        if self._drop_indexes and entity_timestamp_column:
            self._append_drop_column(entity_timestamp_column)

        for feature_set in feature_set_objects.values():
            if self._drop_indexes:
                self._append_drop_column(feature_set.spec.timestamp_key)
            for key in feature_set.spec.entities.keys():
                self._append_index(key)

        start_time = str_to_timestamp(start_time)
        end_time = str_to_timestamp(end_time)
        if start_time and not end_time:
            # if end_time is not specified set it to now()
            end_time = pd.Timestamp.now()

        return self._generate_offline_vector(
            entity_rows,
            entity_timestamp_column,
            feature_set_objects=feature_set_objects,
            feature_set_fields=feature_set_fields,
            start_time=start_time,
            end_time=end_time,
            timestamp_for_filtering=timestamp_for_filtering,
            query=query,
            order_by=order_by,
            additional_filters=additional_filters,
        )

    def _write_to_offline_target(self, timestamp_key=None):
        save_vector = False
        if not self._drop_indexes and timestamp_key not in self._drop_columns:
            self.vector.status.timestamp_key = timestamp_key
            save_vector = True
        if self._target:
            is_persistent_vector = self.vector.metadata.name is not None
            if not self._target.path and not is_persistent_vector:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "target path was not specified"
                )
            self._target.set_resource(self.vector)
            size = self._target.write_dataframe(
                self._result_df, timestamp_key=self.vector.status.timestamp_key
            )
            if is_persistent_vector:
                target_status = self._target.update_resource_status("ready", size=size)
                logger.info(f"wrote target: {target_status}")
            save_vector = True
        if save_vector:
            self.vector.save()

    def _set_indexes(self, df):
        if self._index_columns and not self._drop_indexes:
            if df.index is None or df.index.name is None:
                index_columns_missing = []
                for index in self._index_columns:
                    if index not in df.columns:
                        index_columns_missing.append(index)
                if not index_columns_missing:
                    df.set_index(self._index_columns, inplace=True)
                else:
                    logger.warn(
                        f"Can't set index, not all index columns found: {index_columns_missing}. "
                        f"It is possible that column was already indexed."
                    )
        else:
            df.reset_index(drop=True, inplace=True)

    def _generate_offline_vector(
        self,
        entity_rows,
        entity_timestamp_column,
        feature_set_objects,
        feature_set_fields,
        start_time=None,
        end_time=None,
        timestamp_for_filtering=None,
        query=None,
        order_by=None,
        additional_filters=None,
    ):
        self._create_engine_env()

        feature_sets = []
        dfs = []
        keys = []  # the struct of key is [[[],[]], ..] So that each record indicates which way the corresponding
        # featureset is connected to the previous one, and within each record the left keys are indicated in index 0
        # and the right keys in index 1, this keys will be the keys that will be used in this join
        join_types = []

        if entity_rows is not None:
            if entity_rows.index.names[0]:
                entity_rows = entity_rows.reset_index()
            entity_rows_keys = list(entity_rows.columns)
        else:
            entity_rows_keys = None
        join_graph = self._get_graph(
            feature_set_objects, feature_set_fields, entity_rows_keys
        )
        if entity_rows_keys:
            entity_rows = self._convert_entity_rows_to_engine_df(entity_rows)
            dfs.append(entity_rows)
            keys.append([[], []])
            feature_sets.append(None)
            join_types.append(None)

        timestamp_filtered = False
        for step in join_graph.steps:
            name = step.right_feature_set_name
            feature_set = feature_set_objects[name]
            saved_columns_for_relation = list(
                self.vector.get_feature_set_relations(feature_set).keys()
            )
            feature_sets.append(feature_set)
            columns = feature_set_fields[name]
            self._origin_alias.update({name: alias for name, alias in columns})
            column_names = [name for name, _ in columns]

            for column in saved_columns_for_relation:
                if column not in column_names:
                    column_names.append(column)
                    if column not in self._index_columns:
                        self._append_drop_column(column)

            if isinstance(timestamp_for_filtering, dict):
                time_column = timestamp_for_filtering.get(
                    name, feature_set.spec.timestamp_key
                )
            elif isinstance(timestamp_for_filtering, str):
                time_column = timestamp_for_filtering
            else:
                time_column = feature_set.spec.timestamp_key

            if time_column != feature_set.spec.timestamp_key and time_column not in [
                feature.name for feature in feature_set.spec.features
            ]:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Feature set `{name}` "
                    f"does not have a column named `{time_column}` to filter on."
                )

            if self._drop_indexes:
                self._append_drop_column(time_column)
            if (start_time or end_time) and time_column:
                timestamp_filtered = True

            df = self._get_engine_df(
                feature_set,
                name,
                column_names,
                start_time if time_column else None,
                end_time if time_column else None,
                time_column,
                additional_filters,
            )

            fs_entities_and_timestamp = list(feature_set.spec.entities.keys())
            column_names += fs_entities_and_timestamp
            saved_columns_for_relation += fs_entities_and_timestamp
            if feature_set.spec.timestamp_key:
                column_names.append(feature_set.spec.timestamp_key)
                saved_columns_for_relation.append(feature_set.spec.timestamp_key)
                fs_entities_and_timestamp.append(feature_set.spec.timestamp_key)

            # rename columns to be unique for each feature set and select if needed
            rename_col_dict = {
                column: f"{column}_{name}"
                for column in column_names
                if column not in saved_columns_for_relation
            }
            df_temp = self._rename_columns_and_select(
                df,
                rename_col_dict,
                columns=list(set(column_names + fs_entities_and_timestamp)),
            )

            if df_temp is not None:
                df = df_temp
                del df_temp

            dfs.append(df)
            del df

            keys.append([step.left_keys, step.right_keys])
            join_types.append([step.join_type, step.asof_join])

            # update alias according to the unique column name
            new_columns = []
            if not self._drop_indexes:
                new_columns.extend([(ind, ind) for ind in fs_entities_and_timestamp])
            for column, alias in columns:
                if column in rename_col_dict:
                    new_columns.append((rename_col_dict[column], alias or column))
                else:
                    new_columns.append((column, alias))
            self._update_alias(dictionary={name: alias for name, alias in new_columns})

        # None of the feature sets was timestamp filtered as required
        if not timestamp_filtered and (start_time or end_time):
            raise mlrun.errors.MLRunRuntimeError(
                "start_time and end_time can only be provided in conjunction with "
                "a timestamp column, or when the at least one feature_set has a timestamp key"
            )

        # join the feature data frames
        result_timestamp = self.merge(
            entity_timestamp_column=entity_timestamp_column,
            featuresets=feature_sets,
            featureset_dfs=dfs,
            keys=keys,
            join_types=join_types,
        )

        all_columns = None
        if not self._drop_indexes and result_timestamp:
            if result_timestamp not in self._alias.values():
                self._update_alias(key=result_timestamp, val=result_timestamp)
            all_columns = list(self._alias.keys())

        df_temp = self._rename_columns_and_select(
            self._result_df, self._alias, columns=all_columns
        )
        if df_temp is not None:
            self._result_df = df_temp
            del df_temp

        df_temp = self._drop_columns_from_result()
        if df_temp is not None:
            self._result_df = df_temp
            del df_temp

        if self.vector.status.label_column:
            self._result_df = self._result_df.dropna(
                subset=[self.vector.status.label_column]
            )
        # filter joined data frame by the query param
        if query:
            self._filter(query)

        if order_by:
            if isinstance(order_by, str):
                order_by = [order_by]
            order_by_active = [
                order_col
                if order_col in self._result_df.columns
                else self._origin_alias.get(order_col, None)
                for order_col in order_by
            ]
            if None in order_by_active:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Result dataframe contains {self._result_df.columns} "
                    f"columns and can't order by {order_by}"
                )
            self._order_by(order_by_active)

        self._write_to_offline_target(timestamp_key=result_timestamp)
        return OfflineVectorResponse(self)

    def init_online_vector_service(
        self, entity_keys, fixed_window_type, update_stats=False
    ):
        """
        initialize the `OnlineVectorService`

        :param entity_keys:         list of the feature_vector indexes.
        :param fixed_window_type:   determines how to query the fixed window values which were previously
                                    inserted by ingest
        :param update_stats:        update features statistics from the requested feature sets on the vector.
                                    Default: False.

        :return:                    `OnlineVectorService`
        """
        raise NotImplementedError

    def _unpersist_df(self, df):
        pass

    def _normalize_timestamp_column(
        self,
        entity_timestamp_column,
        reference_df,
        featureset_timestamp,
        featureset_df,
        featureset_name,
    ):
        reference_df_timestamp_type = reference_df[entity_timestamp_column].dtype.name
        featureset_df_timestamp_type = featureset_df[featureset_timestamp].dtype.name

        if reference_df_timestamp_type != featureset_df_timestamp_type:
            logger.info(
                f"Merger detected timestamp resolution incompatibility between feature set {featureset_name} and "
                f"others: {reference_df_timestamp_type} and {featureset_df_timestamp_type}. Converting feature set "
                f"timestamp column '{featureset_timestamp}' to type {reference_df_timestamp_type}."
            )
            featureset_df[featureset_timestamp] = featureset_df[
                featureset_timestamp
            ].astype(reference_df_timestamp_type)

        return featureset_df

    def merge(
        self,
        entity_timestamp_column: str,
        featuresets: list,
        featureset_dfs: list,
        keys: typing.Optional[list] = None,
        join_types: typing.Optional[list] = None,
    ):
        """join the entities and feature set features into a result dataframe"""

        merged_df = featureset_dfs.pop(0)
        featureset = featuresets.pop(0)
        keys.pop(0)
        join_types.pop(0)

        if not entity_timestamp_column and featureset:
            entity_timestamp_column = featureset.spec.timestamp_key

        for featureset, featureset_df, lr_key, join_type in zip(
            featuresets, featureset_dfs, keys, join_types
        ):
            join_type, as_of = join_type
            if (
                featureset.spec.timestamp_key
                and entity_timestamp_column
                and join_type == self._default_join_type
            ):
                merge_func = self._asof_join
            elif join_type == self._default_join_type:
                merge_func = self._join
            elif join_type != self._default_join_type and not as_of:
                self._join_type = join_type
                merge_func = self._join
            else:
                self._join_type = join_type
                merge_func = self._asof_join

            merged_df = merge_func(
                merged_df,
                entity_timestamp_column,
                featureset.metadata.name,
                featureset.spec.timestamp_key,
                featureset_df,
                lr_key[0],
                lr_key[1],
            )
            entity_timestamp_column = (
                entity_timestamp_column or featureset.spec.timestamp_key
            )

            # unpersist as required by the implementation (e.g. spark) and delete references
            # to dataframe to allow for GC to free up the memory (local, dask)
            self._unpersist_df(featureset_df)
            del featureset_df

        self._result_df = merged_df
        return entity_timestamp_column

    def _asof_join(
        self,
        entity_df,
        entity_timestamp_column: str,
        featureset_name: str,
        featureset_timstamp: str,
        featureset_df,
        left_keys: list,
        right_keys: list,
    ):
        raise NotImplementedError("_asof_join() operation not implemented in class")

    def _join(
        self,
        entity_df,
        entity_timestamp_column: str,
        featureset_name: str,
        featureset_timestamp: str,
        featureset_df,
        left_keys: list,
        right_keys: list,
    ):
        raise NotImplementedError("_join() operation not implemented in class")

    def get_status(self):
        """return the status of the merge operation (in case its asynchrounious)"""
        if self._result_df is None:
            raise RuntimeError("unexpected status, no result df")
        return "completed"

    def get_df(self, to_pandas=True):
        """return the result as a dataframe (pandas by default)"""
        self._set_indexes(self._result_df)
        return self._result_df

    def to_parquet(self, target_path, **kw):
        """return results as parquet file"""
        size = ParquetTarget(path=target_path).write_dataframe(self._result_df, **kw)
        return size

    def to_csv(self, target_path, **kw):
        """return results as csv file"""
        size = CSVTarget(path=target_path).write_dataframe(self._result_df, **kw)
        return size

    def _get_graph(
        self, feature_set_objects, feature_set_fields, entity_rows_keys=None
    ):
        join_graph = self.vector.spec.join_graph
        if not join_graph:
            fs_link_list = self._create_linked_relation_list(
                feature_set_objects, feature_set_fields, entity_rows_keys
            )
            join_graph = None
            for i, node in enumerate(fs_link_list):
                if node.name != self._entity_rows_node_name and join_graph is None:
                    join_graph = JoinGraph(first_feature_set=node.name)
                elif node.name == self._entity_rows_node_name:
                    continue
                else:
                    join_graph.inner(other_operand=node.name)

                last_step = join_graph.steps[-1]
                last_step.join_type = self._default_join_type
                last_step.left_keys = node.left_keys
                last_step.right_keys = node.right_keys
        else:
            join_graph._init_all_join_keys(
                feature_set_objects, self.vector, entity_rows_keys
            )
        return join_graph

    class _Node:
        def __init__(
            self,
            name: str,
            order: int,
            left_keys: typing.Optional[list[str]] = None,
            right_keys: typing.Optional[list[str]] = None,
        ):
            self.name = name
            self.left_keys = left_keys if left_keys is not None else []
            self.right_keys = right_keys if right_keys is not None else []
            # order of this feature_set in the original list
            self.order = order
            self.next = None

        def __repr__(self):
            return self.name

        def __eq__(self, other):
            return self.name == other.name

        def __copy__(self):
            return BaseMerger._Node(
                self.name, self.order, self.left_keys, self.right_keys
            )

    class _LinkedList:
        def __init__(self, head=None):
            self.head = head
            self.len = 1 if head is not None else 0

        def __repr__(self):
            node = self.head
            nodes = []
            while node is not None:
                nodes.append(node.name)
                node = node.next
            nodes.append("None")
            return " -> ".join(nodes)

        def __iter__(self):
            node = self.head
            while node is not None:
                yield node
                node = node.next

        def __copy__(self):
            ll = BaseMerger._LinkedList()
            prev_node = None
            for node in self:
                new_node = node.__copy__()
                if ll.head is None:
                    ll.head = new_node
                else:
                    prev_node.next = new_node
                prev_node = new_node
            ll.len = self.len
            return ll

        def add_first(self, node):
            node.next = self.head
            self.head = node
            self.len += 1

        def add_last(self, node):
            if self.head is None:
                self.head = node
                return
            for current_node in self:
                pass
            current_node.next = node
            while node:
                self.len += 1
                node = node.next

        def add_after(self, target_node, new_node):
            new_node.next = target_node.next
            target_node.next = new_node
            self.len += 1

        def find_node(self, target_node_name: str):
            if self.head is None:
                return None

            for node in self:
                if node.name == target_node_name:
                    return node

        def concat(self, other):
            other_iter = iter(other)
            other_head = next(other_iter)
            node = self.find_node(other_head.name)
            if node is None:
                return
            for other_node in other_iter:
                if self.find_node(other_node.name) is None:
                    while node is not None and other_node.order > node.order:
                        node = node.next
                    if node:
                        self.add_after(node, other_node)
                    else:
                        self.add_last(other_node)
                    node = other_node

    def _create_linked_relation_list(
        self, feature_set_objects, feature_set_fields, entity_rows_keys=None
    ):
        feature_set_names = list(feature_set_fields.keys())
        if len(feature_set_names) == 1 and not entity_rows_keys:
            return BaseMerger._LinkedList(
                head=BaseMerger._Node(
                    name=feature_set_names[0],
                    order=0,
                )
            )
        relation_linked_lists = []

        def _create_relation(name: str, order):
            relations = BaseMerger._LinkedList()
            main_node = BaseMerger._Node(
                name,
                order=order,
            )
            relations.add_first(main_node)
            return relations

        def _build_relation(
            fs_name_in: str, name_in_order, linked_list_relation, head_order
        ):
            name_head = linked_list_relation.head.name
            left_keys = feature_set_objects[name_head].extract_relation_keys(
                feature_set_objects[fs_name_in],
                self.vector.get_feature_set_relations(feature_set_objects[name_head]),
            )
            if left_keys == list(
                feature_set_objects[name_head].spec.entities.keys()
            ) and not (head_order < name_in_order):
                left_keys = []

            if left_keys:
                linked_list_relation.add_last(
                    BaseMerger._Node(
                        fs_name_in,
                        left_keys=left_keys,
                        right_keys=list(
                            feature_set_objects[fs_name_in].spec.entities.keys()
                        ),
                        order=name_in_order,
                    )
                )

            return linked_list_relation

        def _build_entity_rows_relation(entity_rows_relation, fs_name, fs_order):
            keys = None
            if feature_set_objects[fs_name].is_connectable_to_df(entity_rows_keys):
                keys = list(feature_set_objects[fs_name].spec.entities.keys())
            if keys:
                entity_rows_relation.add_last(
                    BaseMerger._Node(
                        fs_name,
                        left_keys=keys,
                        right_keys=keys,
                        order=fs_order,
                    )
                )
            return entity_rows_relation

        if entity_rows_keys is not None:
            entity_rows_linked_relation = _create_relation(
                self._entity_rows_node_name, -1
            )
            relation_linked_lists.append(entity_rows_linked_relation)
            linked_list_len_goal = len(feature_set_objects) + 1
        else:
            entity_rows_linked_relation = None
            linked_list_len_goal = len(feature_set_objects)

        for i, name in enumerate(feature_set_names):
            linked_relation = _create_relation(name, i)
            if entity_rows_linked_relation is not None:
                entity_rows_linked_relation = _build_entity_rows_relation(
                    entity_rows_linked_relation, name, i
                )
            for j, name_in in enumerate(feature_set_names):
                if name != name_in:
                    linked_relation = _build_relation(name_in, j, linked_relation, i)
            relation_linked_lists.append(linked_relation)

        # concat all the link lists to one, for the merging process
        for i in range(len(relation_linked_lists)):
            return_relation = relation_linked_lists[i].__copy__()
            for relation_list in relation_linked_lists:
                return_relation.concat(relation_list)
            if return_relation.len == linked_list_len_goal:
                return return_relation

        raise mlrun.errors.MLRunRuntimeError("Failed to merge")

    @classmethod
    def get_default_image(cls, kind):
        return mlrun.mlconf.feature_store.default_job_image

    def _reset_index(self, _result_df):
        raise NotImplementedError

    def _create_engine_env(self):
        """
        initialize engine env if needed
        """
        raise NotImplementedError

    def _get_engine_df(
        self,
        feature_set: FeatureSet,
        feature_set_name: list[str],
        column_names: typing.Optional[list[str]] = None,
        start_time: typing.Optional[typing.Union[str, datetime]] = None,
        end_time: typing.Optional[typing.Union[str, datetime]] = None,
        time_column: typing.Optional[str] = None,
        additional_filters=None,
    ):
        """
        Return the feature_set data frame according to the args

        :param feature_set:             current feature_set to extract from the data frame
        :param feature_set_name:        the name of the current feature_set
        :param column_names:            list of columns to select (if not all)
        :param start_time:              filter by start time
        :param end_time:                filter by end time
        :param time_column:             specify the time column name to filter on

        :return: Data frame of the current engine
        """
        raise NotImplementedError

    def _rename_columns_and_select(
        self,
        df,
        rename_col_dict: dict[str, str],
        columns: typing.Optional[list[str]] = None,
    ):
        """
        rename the columns of the df according to rename_col_dict, and select only `columns` if it is not none

        :param df:              the data frame to change
        :param rename_col_dict: the renaming dictionary - {<current_column_name>: <new_column_name>, ...}
        :param columns:         list of columns to select (if not all)

        :return: the data frame after the transformation or None if the transformation were preformed inplace
        """
        raise NotImplementedError

    def _drop_columns_from_result(self):
        """
        drop `self._drop_columns` from `self._result_df`
        """
        raise NotImplementedError

    def _filter(self, query: str):
        """
        filter `self._result_df` by `query`

        :param query: The query string used to filter rows
        """
        raise NotImplementedError

    def _order_by(self, order_by_active: list[str]):
        """
        Order by `order_by_active` along all axis.

        :param order_by_active: list of names to sort by.
        """
        raise NotImplementedError

    def _convert_entity_rows_to_engine_df(self, entity_rows):
        raise NotImplementedError



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib
import typing

import nuclio
import nuclio.auth

import mlrun.common.schemas as schemas
import mlrun.errors
import mlrun.run
from mlrun.common.runtimes.constants import NuclioIngressAddTemplatedIngressModes
from mlrun.runtimes import RemoteRuntime
from mlrun.runtimes.nuclio import (
    min_nuclio_versions,
    multiple_port_sidecar_is_supported,
)
from mlrun.runtimes.nuclio.api_gateway import (
    APIGateway,
    APIGatewayMetadata,
    APIGatewaySpec,
)
from mlrun.runtimes.nuclio.function import NuclioSpec, NuclioStatus
from mlrun.utils import is_valid_port, logger, update_in


class ApplicationSpec(NuclioSpec):
    _dict_fields = NuclioSpec._dict_fields + [
        "internal_application_port",
        "application_ports",
    ]

    def __init__(
        self,
        command=None,
        args=None,
        image=None,
        mode=None,
        entry_points=None,
        description=None,
        replicas=None,
        min_replicas=None,
        max_replicas=None,
        volumes=None,
        volume_mounts=None,
        env=None,
        resources=None,
        config=None,
        base_spec=None,
        no_cache=None,
        source=None,
        image_pull_policy=None,
        function_kind=None,
        build=None,
        service_account=None,
        readiness_timeout=None,
        readiness_timeout_before_failure=None,
        default_handler=None,
        node_name=None,
        node_selector=None,
        affinity=None,
        disable_auto_mount=False,
        priority_class_name=None,
        pythonpath=None,
        workdir=None,
        image_pull_secret=None,
        tolerations=None,
        preemption_mode=None,
        security_context=None,
        service_type=None,
        add_templated_ingress_host_mode=None,
        state_thresholds=None,
        disable_default_http_trigger=None,
        serving_spec=None,
        graph=None,
        parameters=None,
        track_models=None,
        internal_application_port=None,
        application_ports=None,
    ):
        super().__init__(
            command=command,
            args=args,
            image=image,
            mode=mode,
            entry_points=entry_points,
            description=description,
            replicas=replicas,
            min_replicas=min_replicas,
            max_replicas=max_replicas,
            volumes=volumes,
            volume_mounts=volume_mounts,
            env=env,
            resources=resources,
            config=config,
            base_spec=base_spec,
            no_cache=no_cache,
            source=source,
            image_pull_policy=image_pull_policy,
            function_kind=function_kind,
            build=build,
            service_account=service_account,
            readiness_timeout=readiness_timeout,
            readiness_timeout_before_failure=readiness_timeout_before_failure,
            default_handler=default_handler,
            node_name=node_name,
            node_selector=node_selector,
            affinity=affinity,
            disable_auto_mount=disable_auto_mount,
            priority_class_name=priority_class_name,
            pythonpath=pythonpath,
            workdir=workdir,
            image_pull_secret=image_pull_secret,
            tolerations=tolerations,
            preemption_mode=preemption_mode,
            security_context=security_context,
            service_type=service_type,
            add_templated_ingress_host_mode=add_templated_ingress_host_mode,
            serving_spec=serving_spec,
            graph=graph,
            parameters=parameters,
            track_models=track_models,
            state_thresholds=state_thresholds,
            disable_default_http_trigger=disable_default_http_trigger,
        )

        # Override default min/max replicas (don't assume application is stateless)
        self.min_replicas = min_replicas or 1
        self.max_replicas = max_replicas or 1

        # initializing internal application port and application ports
        self._internal_application_port = None
        self._application_ports = []

        application_ports = application_ports or []

        # if internal_application_port is not provided, use the first application port
        if not internal_application_port and len(application_ports) > 0:
            internal_application_port = application_ports[0]

        # the port of application sidecar to which traffic will be routed from a nuclio function
        self.internal_application_port = (
            internal_application_port
            or mlrun.mlconf.function.application.default_sidecar_internal_port
        )

        # all exposed ports by the application sidecar
        self.application_ports = application_ports

    @property
    def application_ports(self):
        return self._application_ports

    @application_ports.setter
    def application_ports(self, ports):
        """
        Set the application ports for the application sidecar.
        The internal application port is always included and always first.
        """
        # Handle None / single int
        if ports is None:
            ports = []
        elif isinstance(ports, int):
            ports = [ports]
        elif not isinstance(ports, list):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Application ports must be a list of integers"
            )

        # Validate and normalize
        cleaned_ports = []
        for port in ports:
            is_valid_port(port, raise_on_error=True)
            if port != self.internal_application_port:
                cleaned_ports.append(port)

        application_ports = [self.internal_application_port] + cleaned_ports

        # ensure multiple ports are supported in Nuclio
        if len(application_ports) > 1:
            multiple_port_sidecar_is_supported()

        self._application_ports = application_ports

    @property
    def internal_application_port(self):
        return self._internal_application_port

    @internal_application_port.setter
    def internal_application_port(self, port):
        port = int(port)
        is_valid_port(port, raise_on_error=True)
        self._internal_application_port = port

        # If when internal application port is being set, length of self._application_ports is 1,
        # it means that it consist of [old_port] only
        # so in this case, we rewrite the list completely, by setting value to [new_value]
        if len(self.application_ports) == 1:
            self._application_ports = [port]
            return

        # when setting new internal application port, ensure that it is included in the application ports
        # it just triggers setter logic, so setting to the same value is a no-op
        self.application_ports = self._application_ports


class ApplicationStatus(NuclioStatus):
    def __init__(
        self,
        state=None,
        nuclio_name=None,
        address=None,
        internal_invocation_urls=None,
        external_invocation_urls=None,
        build_pod=None,
        container_image=None,
        application_image=None,
        application_source=None,
        sidecar_name=None,
        api_gateway_name=None,
        api_gateway=None,
        url=None,
    ):
        super().__init__(
            state=state,
            nuclio_name=nuclio_name,
            address=address,
            internal_invocation_urls=internal_invocation_urls,
            external_invocation_urls=external_invocation_urls,
            build_pod=build_pod,
            container_image=container_image,
        )
        self.application_image = application_image or None
        self.application_source = application_source or None
        self.sidecar_name = sidecar_name or None
        self.api_gateway_name = api_gateway_name or None
        self.api_gateway: typing.Optional[APIGateway] = api_gateway or None
        self.url = url or None


class ApplicationRuntime(RemoteRuntime):
    kind = "application"
    reverse_proxy_image = None

    @min_nuclio_versions("1.13.1")
    def __init__(self, spec=None, metadata=None):
        super().__init__(spec=spec, metadata=metadata)

    @property
    def spec(self) -> ApplicationSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", ApplicationSpec)

    @property
    def status(self) -> ApplicationStatus:
        return self._status

    @status.setter
    def status(self, status):
        self._status = self._verify_dict(status, "status", ApplicationStatus)

    @property
    def api_gateway(self):
        return self.status.api_gateway

    @api_gateway.setter
    def api_gateway(self, api_gateway: APIGateway):
        self.status.api_gateway = api_gateway

    @property
    def url(self):
        if not self.status.api_gateway:
            self._sync_api_gateway()
        return self.status.api_gateway.invoke_url

    @url.setter
    def url(self, url):
        self.status.url = url

    def set_internal_application_port(self, port: int):
        self.spec.internal_application_port = port

    def with_sidecar(
        self,
        name: typing.Optional[str] = None,
        image: typing.Optional[str] = None,
        ports: typing.Optional[typing.Union[int, list[int]]] = None,
        command: typing.Optional[str] = None,
        args: typing.Optional[list[str]] = None,
    ):
        # wraps with_sidecar just to set the application ports
        super().with_sidecar(
            name=name,
            image=image,
            ports=ports,
            command=command,
            args=args,
        )

        if ports:
            if self.spec.internal_application_port != ports[0]:
                logger.info(
                    f"Setting internal application port to the first port from the sidecar: {ports[0]}. "
                    f"If this is not intended, please set the internal_application_port explicitly."
                )
                self.spec.internal_application_port = ports[0]
            self.spec.application_ports = ports

    def pre_deploy_validation(self):
        super().pre_deploy_validation()
        if not self.spec.config.get("spec.sidecars"):
            raise mlrun.errors.MLRunBadRequestError(
                "Application spec must include a sidecar configuration"
            )

        sidecars = self.spec.config["spec.sidecars"]
        for sidecar in sidecars:
            if not sidecar.get("image"):
                raise mlrun.errors.MLRunBadRequestError(
                    "Application sidecar spec must include an image"
                )

            if not sidecar.get("ports"):
                raise mlrun.errors.MLRunBadRequestError(
                    "Application sidecar spec must include at least one port"
                )

            ports = sidecar["ports"]
            for port in ports:
                if not port.get("containerPort"):
                    raise mlrun.errors.MLRunBadRequestError(
                        "Application sidecar port spec must include a containerPort"
                    )

                if not port.get("name"):
                    raise mlrun.errors.MLRunBadRequestError(
                        "Application sidecar port spec must include a name"
                    )

            if not sidecar.get("command") and sidecar.get("args"):
                raise mlrun.errors.MLRunBadRequestError(
                    "Application sidecar spec must include a command if args are provided"
                )

    def prepare_image_for_deploy(self):
        if self.spec.build.source and self.spec.build.load_source_on_run:
            logger.warning(
                "Application runtime requires loading the source into the application image. "
                f"Even though {self.spec.build.load_source_on_run=}, loading on build will be forced."
            )
            self.spec.build.load_source_on_run = False
        super().prepare_image_for_deploy()

    def deploy(
        self,
        project="",
        tag="",
        verbose=False,
        builder_env: typing.Optional[dict] = None,
        force_build: bool = False,
        with_mlrun=None,
        skip_deployed=False,
        is_kfp=False,
        mlrun_version_specifier=None,
        show_on_failure: bool = False,
        create_default_api_gateway: bool = True,
    ):
        """
        Deploy function, builds the application image if required (self.requires_build()) or force_build is True,
        Once the image is built, the function is deployed.

        :param project:                     Project name
        :param tag:                         Function tag
        :param verbose:                     Set True for verbose logging
        :param builder_env:                 Env vars dict for source archive config/credentials
                                            e.g. builder_env={"GIT_TOKEN": token}
        :param force_build:                 Set True for force building the application image
        :param with_mlrun:                  Add the current mlrun package to the container build
        :param skip_deployed:               Skip the build if we already have an image for the function
        :param is_kfp:                      Deploy as part of a kfp pipeline
        :param mlrun_version_specifier:     Which mlrun package version to include (if not current)
        :param show_on_failure:             Show logs only in case of build failure
        :param create_default_api_gateway:  When deploy finishes the default API gateway will be created for the
                                            application. Disabling this flag means that the application will not be
                                            accessible until an API gateway is created for it.

        :return: The default API gateway URL if created or True if the function is ready (deployed)
        """
        mlrun.utils.helpers.validate_function_name(self.metadata.name)

        if (self.requires_build() and not self.spec.image) or force_build:
            self._fill_credentials()
            self._build_application_image(
                builder_env=builder_env,
                force_build=force_build,
                watch=True,
                with_mlrun=with_mlrun,
                skip_deployed=skip_deployed,
                is_kfp=is_kfp,
                mlrun_version_specifier=mlrun_version_specifier,
                show_on_failure=show_on_failure,
            )

        self._ensure_reverse_proxy_configurations()
        self._configure_application_sidecar()

        # We only allow accessing the application via the API Gateway
        self.spec.add_templated_ingress_host_mode = (
            NuclioIngressAddTemplatedIngressModes.never
        )

        super().deploy(
            project=project,
            tag=tag,
            verbose=verbose,
            builder_env=builder_env,
        )
        logger.info(
            "Successfully deployed function.",
        )

        # Restore the source in case it was removed to make nuclio not consider it when building
        if not self.spec.build.source and self.status.application_source:
            self.spec.build.source = self.status.application_source
        self.save(versioned=False)

        if create_default_api_gateway:
            try:
                api_gateway_name = self.resolve_default_api_gateway_name()
                return self.create_api_gateway(api_gateway_name, set_as_default=True)
            except Exception as exc:
                logger.warning(
                    "Failed to create default API gateway, application may not be accessible. "
                    "Use the `create_api_gateway` method to make it accessible",
                    exc=mlrun.errors.err_to_str(exc),
                )
        elif not self.status.api_gateway:
            logger.warning(
                "Application is online but may not be accessible since default gateway creation was not requested."
                "Use the `create_api_gateway` method to make it accessible."
            )

        return True

    def with_source_archive(
        self,
        source,
        workdir=None,
        pull_at_runtime: bool = False,
        target_dir: typing.Optional[str] = None,
    ):
        """load the code from git/tar/zip archive at build

        :param source:          valid absolute path or URL to git, zip, or tar file, e.g.
                                git://github.com/mlrun/something.git
                                http://some/url/file.zip
                                note path source must exist on the image or exist locally when run is local
                                (it is recommended to use 'workdir' when source is a filepath instead)
        :param workdir:         working dir relative to the archive root (e.g. './subdir') or absolute to the image root
        :param pull_at_runtime: currently not supported, source must be loaded into the image during the build process
        :param target_dir:      target dir on runtime pod or repo clone / archive extraction
        """
        if pull_at_runtime:
            logger.warning(
                f"{pull_at_runtime=} is currently not supported for application runtime "
                "and will be overridden to False",
                pull_at_runtime=pull_at_runtime,
            )

        self._configure_mlrun_build_with_source(
            source=source,
            workdir=workdir,
            pull_at_runtime=False,
            target_dir=target_dir,
        )

    def from_image(self, image):
        """
        Deploy the function with an existing nuclio processor image.
        This applies only for the reverse proxy and not the application image.

        :param image: image name
        """
        super().from_image(image)
        # nuclio implementation detail - when providing the image and emptying out the source code and build source,
        # nuclio skips rebuilding the image and simply takes the prebuilt image
        self.spec.build.functionSourceCode = ""
        self.spec.config.pop("spec.build.functionSourceCode", None)
        self.status.application_source = self.spec.build.source
        self.spec.build.source = ""
        self.spec.config.pop("spec.build.source", None)

        # save the image in the status, so we won't repopulate the function source code
        self.status.container_image = image

        # ensure golang runtime and handler for the reverse proxy
        self.spec.nuclio_runtime = "golang"
        update_in(
            self.spec.base_spec,
            "spec.handler",
            "main:Handler",
        )

    @staticmethod
    def get_filename_and_handler() -> (str, str):
        reverse_proxy_file_path = pathlib.Path(__file__).parent / "reverse_proxy.go"
        return str(reverse_proxy_file_path), "Handler"

    def create_api_gateway(
        self,
        name: typing.Optional[str] = None,
        path: typing.Optional[str] = None,
        direct_port_access: bool = False,
        authentication_mode: schemas.APIGatewayAuthenticationMode = None,
        authentication_creds: typing.Optional[tuple[str, str]] = None,
        ssl_redirect: typing.Optional[bool] = None,
        set_as_default: bool = False,
        gateway_timeout: typing.Optional[int] = None,
        port: typing.Optional[int] = None,
    ):
        """
        Create the application API gateway. Once the application is deployed, the API gateway can be created.
        An application without an API gateway is not accessible.

        :param name:                    The name of the API gateway
        :param path:                    Optional path of the API gateway, default value is "/".
                                        The given path should be supported by the deployed application
        :param direct_port_access:      Set True to allow direct port access to the application sidecar
        :param authentication_mode:     API Gateway authentication mode
        :param authentication_creds:    API Gateway basic authentication credentials as a tuple (username, password)
        :param ssl_redirect:            Set True to force SSL redirect, False to disable. Defaults to
                                        mlrun.mlconf.force_api_gateway_ssl_redirect()
        :param set_as_default:          Set the API gateway as the default for the application (`status.api_gateway`)
        :param gateway_timeout:         nginx ingress timeout in sec (request timeout, when will the gateway return an
                                        error)
        :param port:                    The API gateway port, used only when direct_port_access=True

        :return:                        The API gateway URL
        """
        if not name:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "API gateway name must be specified."
            )

        if not set_as_default and name == self.resolve_default_api_gateway_name():
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Non-default API gateway cannot use the default gateway name, {name=}."
            )

        if (
            authentication_mode == schemas.APIGatewayAuthenticationMode.basic
            and not authentication_creds
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Authentication credentials not provided"
            )

        if direct_port_access and port:
            logger.warning(
                "Ignoring 'port' because 'direct_port_access' is enabled. "
                "The 'port' setting is only applicable when 'direct_port_access' is disabled."
            )

        ports = (
            port or self.spec.internal_application_port if direct_port_access else []
        )

        api_gateway = APIGateway(
            APIGatewayMetadata(
                name=name,
                namespace=self.metadata.namespace,
                labels=self.metadata.labels.copy(),
            ),
            APIGatewaySpec(
                functions=[self],
                project=self.metadata.project,
                path=path,
                ports=mlrun.utils.helpers.as_list(ports) if ports else None,
            ),
        )

        api_gateway.with_gateway_timeout(gateway_timeout)
        if ssl_redirect is None:
            ssl_redirect = mlrun.mlconf.force_api_gateway_ssl_redirect()
        if ssl_redirect:
            # Force ssl redirect so that the application is only accessible via https
            api_gateway.with_force_ssl_redirect()

        # Add authentication if required
        authentication_mode = (
            authentication_mode
            or mlrun.mlconf.function.application.default_authentication_mode
        )
        if authentication_mode == schemas.APIGatewayAuthenticationMode.access_key:
            api_gateway.with_access_key_auth()
        elif authentication_mode == schemas.APIGatewayAuthenticationMode.basic:
            api_gateway.with_basic_auth(*authentication_creds)

        db = self._get_db()
        api_gateway_scheme = db.store_api_gateway(
            api_gateway=api_gateway.to_scheme(), project=self.metadata.project
        )

        if set_as_default:
            self.status.api_gateway_name = api_gateway_scheme.metadata.name
            self.status.api_gateway = APIGateway.from_scheme(api_gateway_scheme)
            self.status.api_gateway.wait_for_readiness()
            self.url = self.status.api_gateway.invoke_url
            url = self.url
        else:
            api_gateway = APIGateway.from_scheme(api_gateway_scheme)
            api_gateway.wait_for_readiness()
            url = api_gateway.invoke_url
            # Update application status (enriches invocation url)
            self._get_state(raise_on_exception=False)

        logger.info("Successfully created API gateway", url=url)
        return url

    def delete_api_gateway(self, name: str):
        """
        Delete API gateway by name.
        Refreshes the application status to update api gateway and invocation URLs.
        :param name:    The API gateway name
        """
        self._get_db().delete_api_gateway(name=name, project=self.metadata.project)
        if name == self.status.api_gateway_name:
            self.status.api_gateway_name = None
            self.status.api_gateway = None
        self._get_state()

    def invoke(
        self,
        path: str = "",
        body: typing.Optional[typing.Union[str, bytes, dict]] = None,
        method: typing.Optional[str] = None,
        headers: typing.Optional[dict] = None,
        force_external_address: bool = False,
        auth_info: schemas.AuthInfo = None,
        mock: typing.Optional[bool] = None,
        credentials: typing.Optional[tuple[str, str]] = None,
        **http_client_kwargs,
    ):
        self._sync_api_gateway()

        # If the API Gateway is not ready or not set, try to invoke the function directly (without the API Gateway)
        if not self.status.api_gateway:
            logger.warning(
                "Default API gateway is not configured, invoking function invocation URL."
            )
            # create a requests auth object if credentials are provided and not already set in the http client kwargs
            auth = http_client_kwargs.pop("auth", None) or (
                nuclio.auth.AuthInfo(
                    username=credentials[0], password=credentials[1]
                ).to_requests_auth()
                if credentials
                else None
            )
            return super().invoke(
                path,
                body,
                method,
                headers,
                force_external_address,
                auth_info,
                mock,
                auth=auth,
                **http_client_kwargs,
            )

        if not method:
            method = "POST" if body else "GET"

        return self.status.api_gateway.invoke(
            method=method,
            headers=headers,
            credentials=credentials,
            path=path,
            body=body,
            **http_client_kwargs,
        )

    @classmethod
    def deploy_reverse_proxy_image(cls):
        """
        Build the reverse proxy image and save it.
        The reverse proxy image is used to route requests to the application sidecar.
        This is useful when you want to decrease build time by building the application image only once.

        :param use_cache:   Use the cache when building the image
        """
        # create a function that includes only the reverse proxy, without the application

        if not mlrun.get_current_project(silent=True):
            raise mlrun.errors.MLRunMissingProjectError(
                "An active project is required to run deploy_reverse_proxy_image(). "
                "Use `mlrun.get_or_create_project()` or set an active project first."
            )

        reverse_proxy_func = mlrun.run.new_function(
            name="reverse-proxy-temp", kind="remote"
        )
        # default max replicas is 4, we only need one replica for the reverse proxy
        reverse_proxy_func.spec.max_replicas = 1

        # the reverse proxy image should not be based on another image
        reverse_proxy_func.set_config("spec.build.baseImage", None)
        reverse_proxy_func.spec.image = ""
        reverse_proxy_func.spec.build.base_image = ""

        cls._ensure_reverse_proxy_configurations(reverse_proxy_func)
        reverse_proxy_func.deploy()

        # save the created container image
        cls.reverse_proxy_image = reverse_proxy_func.status.container_image

        # delete the function to avoid cluttering the project
        mlrun.get_run_db().delete_function(
            reverse_proxy_func.metadata.name, reverse_proxy_func.metadata.project
        )

    def resolve_default_api_gateway_name(self):
        return (
            f"{self.metadata.name}-{self.metadata.tag}"
            if self.metadata.tag
            else self.metadata.name
        )

    @min_nuclio_versions("1.13.1")
    def disable_default_http_trigger(
        self,
    ):
        raise mlrun.runtimes.RunError(
            "Application runtime does not support disabling the default HTTP trigger"
        )

    @min_nuclio_versions("1.13.1")
    def enable_default_http_trigger(
        self,
    ):
        pass

    def _run(self, runobj: "mlrun.RunObject", execution):
        raise mlrun.runtimes.RunError(
            "Application runtime .run() is not yet supported. Use .invoke() instead."
        )

    def _enrich_command_from_status(self):
        pass

    def _build_application_image(
        self,
        builder_env: typing.Optional[dict] = None,
        force_build: bool = False,
        watch=True,
        with_mlrun=None,
        skip_deployed=False,
        is_kfp=False,
        mlrun_version_specifier=None,
        show_on_failure: bool = False,
    ):
        if not self.spec.command:
            logger.warning(
                "Building the application image without a command. "
                "Use spec.command and spec.args to specify the application entrypoint",
                command=self.spec.command,
                args=self.spec.args,
            )

        if self.spec.build.source in [".", "./"]:
            logger.info(
                "The application is configured to use the project's source. "
                "Application runtime requires loading the source into the application image. "
                "Loading on build will be forced regardless of whether 'pull_at_runtime=True' was configured."
            )

        with_mlrun = self._resolve_build_with_mlrun(with_mlrun)
        return self._build_image(
            builder_env=builder_env,
            force_build=force_build,
            mlrun_version_specifier=mlrun_version_specifier,
            show_on_failure=show_on_failure,
            skip_deployed=skip_deployed,
            watch=watch,
            is_kfp=is_kfp,
            with_mlrun=with_mlrun,
        )

    def _ensure_reverse_proxy_configurations(self):
        # If an HTTP trigger already exists in the spec,
        # it means the user explicitly defined a custom configuration,
        # so, skip automatic creation.
        skip_http_trigger_creation = False
        for key, value in self.spec.config.items():
            if key.startswith("spec.triggers"):
                if isinstance(value, dict):
                    if value.get("kind") == "http":
                        skip_http_trigger_creation = True
                        break
        if not skip_http_trigger_creation:
            self.with_http(
                workers=mlrun.mlconf.function.application.default_worker_number,
                trigger_name="application-http",
            )

        if self.spec.build.functionSourceCode or self.status.container_image:
            return

        filename, handler = ApplicationRuntime.get_filename_and_handler()
        name, spec, code = nuclio.build_file(
            filename,
            name=self.metadata.name,
            handler=handler,
        )
        self.spec.function_handler = mlrun.utils.get_in(spec, "spec.handler")
        self.spec.build.functionSourceCode = mlrun.utils.get_in(
            spec, "spec.build.functionSourceCode"
        )
        self.spec.nuclio_runtime = mlrun.utils.get_in(spec, "spec.runtime")

        # default the reverse proxy logger level to info
        logger_sinks_key = "spec.loggerSinks"
        if not self.spec.config.get(logger_sinks_key):
            self.set_config(
                logger_sinks_key, [{"level": "info", "sink": "myStdoutLoggerSink"}]
            )

    def _configure_application_sidecar(self):
        # Save the application image in the status to allow overriding it with the reverse proxy entry point
        if self.spec.image and (
            not self.status.application_image
            or self.spec.image != self.status.container_image
        ):
            self.status.application_image = self.spec.image
            self.spec.image = ""

        # reuse the reverse proxy image if it was built before
        if (
            reverse_proxy_image := self.status.container_image
            or self.reverse_proxy_image
        ):
            self.from_image(reverse_proxy_image)

        self.status.sidecar_name = f"{self.metadata.name}-sidecar"
        self.with_sidecar(
            name=self.status.sidecar_name,
            image=self.status.application_image,
            ports=self.spec.application_ports,
            command=self.spec.command,
            args=self.spec.args,
        )
        self.set_env("SIDECAR_PORT", self.spec.internal_application_port)
        self.set_env("SIDECAR_HOST", "http://localhost")

        # configure the sidecar container as the default container for logging purposes
        self.metadata.annotations["kubectl.kubernetes.io/default-container"] = (
            self.status.sidecar_name
        )

    def _sync_api_gateway(self):
        if not self.status.api_gateway_name:
            return

        db = self._get_db()
        api_gateway_scheme = db.get_api_gateway(
            name=self.status.api_gateway_name, project=self.metadata.project
        )
        self.status.api_gateway = APIGateway.from_scheme(api_gateway_scheme)
        self.status.api_gateway.wait_for_readiness()
        self.url = self.status.api_gateway.invoke_url



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import mlrun.common.constants as mlrun_constants
import mlrun_pipelines.common.models


class PodPhases:
    """
    https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle/#pod-phase
    """

    succeeded = "Succeeded"
    failed = "Failed"
    pending = "Pending"
    running = "Running"
    unknown = "Unknown"

    @staticmethod
    def terminal_phases():
        return [
            PodPhases.succeeded,
            PodPhases.failed,
        ]

    @staticmethod
    def all():
        return [
            PodPhases.succeeded,
            PodPhases.failed,
            PodPhases.pending,
            PodPhases.running,
            PodPhases.unknown,
        ]

    @staticmethod
    def pod_phase_to_run_state(pod_phase):
        if pod_phase not in PodPhases.all():
            raise ValueError(f"Invalid pod phase: {pod_phase}")
        return {
            PodPhases.succeeded: RunStates.completed,
            PodPhases.failed: RunStates.error,
            PodPhases.pending: RunStates.pending,
            PodPhases.running: RunStates.running,
            PodPhases.unknown: RunStates.unknown,
        }[pod_phase]


class ThresholdStates:
    # A pod can be in pending and scheduled state when the pod's container images are not already available on the node
    # where it is scheduled, or initialization tasks specified in the pod's configuration are not yet completed.
    pending_scheduled = "pending_scheduled"
    pending_not_scheduled = "pending_not_scheduled"
    executing = "executing"
    image_pull_backoff = "image_pull_backoff"

    @staticmethod
    def all():
        return [
            ThresholdStates.pending_scheduled,
            ThresholdStates.pending_not_scheduled,
            ThresholdStates.executing,
            ThresholdStates.image_pull_backoff,
        ]

    @staticmethod
    def from_pod_phase(pod_phase: str, pod: dict) -> typing.Optional[str]:
        if pod_phase == PodPhases.pending:
            if ThresholdStates.is_pod_in_image_pull_backoff(pod):
                return ThresholdStates.image_pull_backoff
            elif ThresholdStates.is_pod_scheduled(pod):
                return ThresholdStates.pending_scheduled
            else:
                return ThresholdStates.pending_not_scheduled

        elif pod_phase == PodPhases.running:
            return ThresholdStates.executing

        return None

    @staticmethod
    def is_pod_scheduled(pod: dict):
        conditions = pod["status"].get("conditions", []) or []
        for condition in conditions:
            if condition["type"] == "PodScheduled" and condition["status"] == "True":
                return True
        return False

    @staticmethod
    def is_pod_in_image_pull_backoff(pod: dict):
        container_statuses = pod.get("status").get("container_statuses", []) or []
        for container_status in container_statuses:
            state_waiting = container_status.get("state", {}).get("waiting", {}) or {}
            if state_waiting.get("reason", "") == "ImagePullBackOff":
                return True
        return False


class MPIJobCRDVersions:
    v1 = "v1"
    v1alpha1 = "v1alpha1"

    @staticmethod
    def all():
        return [MPIJobCRDVersions.v1, MPIJobCRDVersions.v1alpha1]

    @staticmethod
    def default():
        return MPIJobCRDVersions.v1alpha1

    @staticmethod
    def role_label_by_version(version):
        return {
            MPIJobCRDVersions.v1alpha1: mlrun_constants.MLRunInternalLabels.mpi_role_type,
            MPIJobCRDVersions.v1: mlrun_constants.MLRunInternalLabels.mpi_job_role,
        }[version]


class RunStates:
    completed = "completed"
    error = "error"
    running = "running"
    created = "created"
    pending = "pending"
    unknown = "unknown"
    aborted = "aborted"
    aborting = "aborting"
    skipped = "skipped"
    pending_retry = "pendingRetry"

    @staticmethod
    def all():
        return [
            RunStates.completed,
            RunStates.error,
            RunStates.running,
            RunStates.created,
            RunStates.pending,
            RunStates.unknown,
            RunStates.aborted,
            RunStates.aborting,
            RunStates.skipped,
            RunStates.pending_retry,
        ]

    @staticmethod
    def terminal_states():
        return [
            RunStates.completed,
            RunStates.error,
            RunStates.aborted,
            RunStates.skipped,
        ]

    @staticmethod
    def error_states():
        return [
            RunStates.error,
            RunStates.aborted,
            RunStates.pending_retry,
        ]

    @staticmethod
    def abortion_states():
        return [
            RunStates.aborted,
            RunStates.aborting,
        ]

    @staticmethod
    def error_and_abortion_states():
        return list(set(RunStates.error_states()) | set(RunStates.abortion_states()))

    @staticmethod
    def non_terminal_states():
        return list(set(RunStates.all()) - set(RunStates.terminal_states()))

    @staticmethod
    def terminal_or_error_states():
        return list(
            set(RunStates.terminal_states())
            | set(RunStates.error_and_abortion_states())
        )

    @staticmethod
    def not_allowed_for_deletion_states():
        return [
            RunStates.running,
            RunStates.pending,
        ]

    @staticmethod
    def notification_states():
        return RunStates.terminal_states() + [RunStates.running]

    @staticmethod
    def run_state_to_pipeline_run_status(run_state: str):
        if not run_state:
            return mlrun_pipelines.common.models.RunStatuses.runtime_state_unspecified

        if run_state not in RunStates.all():
            raise ValueError(f"Invalid run state: {run_state}")

        return {
            RunStates.completed: mlrun_pipelines.common.models.RunStatuses.succeeded,
            RunStates.error: mlrun_pipelines.common.models.RunStatuses.failed,
            RunStates.running: mlrun_pipelines.common.models.RunStatuses.running,
            RunStates.created: mlrun_pipelines.common.models.RunStatuses.pending,
            RunStates.pending: mlrun_pipelines.common.models.RunStatuses.pending,
            RunStates.unknown: mlrun_pipelines.common.models.RunStatuses.runtime_state_unspecified,
            RunStates.aborted: mlrun_pipelines.common.models.RunStatuses.canceled,
            RunStates.aborting: mlrun_pipelines.common.models.RunStatuses.canceling,
            RunStates.skipped: mlrun_pipelines.common.models.RunStatuses.skipped,
        }[run_state]

    @staticmethod
    def pipeline_run_status_to_run_state(pipeline_run_status):
        if pipeline_run_status not in mlrun_pipelines.common.models.RunStatuses.all():
            raise ValueError(f"Invalid pipeline run status: {pipeline_run_status}")
        return {
            mlrun_pipelines.common.models.RunStatuses.succeeded: RunStates.completed,
            mlrun_pipelines.common.models.RunStatuses.failed: RunStates.error,
            mlrun_pipelines.common.models.RunStatuses.running: RunStates.running,
            mlrun_pipelines.common.models.RunStatuses.pending: RunStates.pending,
            mlrun_pipelines.common.models.RunStatuses.canceled: RunStates.aborted,
            mlrun_pipelines.common.models.RunStatuses.canceling: RunStates.aborting,
            mlrun_pipelines.common.models.RunStatuses.skipped: RunStates.skipped,
            mlrun_pipelines.common.models.RunStatuses.runtime_state_unspecified: RunStates.unknown,
            mlrun_pipelines.common.models.RunStatuses.error: RunStates.error,
            mlrun_pipelines.common.models.RunStatuses.paused: RunStates.unknown,
            mlrun_pipelines.common.models.RunStatuses.unknown: RunStates.unknown,
        }[pipeline_run_status]


class SparkApplicationStates:
    """
    https://github.com/GoogleCloudPlatform/spark-on-k8s-operator/blob/master/pkg/apis/sparkoperator.k8s.io/v1beta2/types.go#L321
    """

    completed = "COMPLETED"
    failed = "FAILED"
    submitted = "SUBMITTED"
    running = "RUNNING"
    submission_failed = "SUBMISSION_FAILED"
    pending_rerun = "PENDING_RERUN"
    pending_submission = "PENDING_SUBMISSION"
    invalidating = "INVALIDATING"
    succeeding = "SUCCEEDING"
    failing = "FAILING"
    unknown = "UNKNOWN"

    @staticmethod
    def terminal_states():
        return [
            SparkApplicationStates.completed,
            SparkApplicationStates.failed,
        ]

    @staticmethod
    def all():
        return [
            SparkApplicationStates.completed,
            SparkApplicationStates.failed,
            SparkApplicationStates.submitted,
            SparkApplicationStates.running,
            SparkApplicationStates.submission_failed,
            SparkApplicationStates.pending_rerun,
            SparkApplicationStates.pending_submission,
            SparkApplicationStates.invalidating,
            SparkApplicationStates.succeeding,
            SparkApplicationStates.failing,
            SparkApplicationStates.unknown,
        ]

    @staticmethod
    def spark_application_state_to_run_state(spark_application_state):
        if spark_application_state not in SparkApplicationStates.all():
            raise ValueError(
                f"Invalid spark application state: {spark_application_state}"
            )
        return {
            SparkApplicationStates.completed: RunStates.completed,
            SparkApplicationStates.failed: RunStates.error,
            SparkApplicationStates.submitted: RunStates.running,
            SparkApplicationStates.running: RunStates.running,
            SparkApplicationStates.submission_failed: RunStates.running,
            SparkApplicationStates.pending_rerun: RunStates.running,
            SparkApplicationStates.pending_submission: RunStates.running,
            SparkApplicationStates.invalidating: RunStates.running,
            SparkApplicationStates.succeeding: RunStates.running,
            SparkApplicationStates.failing: RunStates.running,
            SparkApplicationStates.unknown: RunStates.unknown,
        }[spark_application_state]


class MPIJobV1Alpha1States:
    """
    https://github.com/kubeflow/mpi-operator/blob/master/pkg/apis/kubeflow/v1alpha1/types.go#L105
    """

    succeeded = "Succeeded"
    failed = "Failed"
    active = "Active"

    @staticmethod
    def terminal_states():
        return [
            MPIJobV1Alpha1States.succeeded,
            MPIJobV1Alpha1States.failed,
        ]

    @staticmethod
    def all():
        return [
            MPIJobV1Alpha1States.succeeded,
            MPIJobV1Alpha1States.failed,
            MPIJobV1Alpha1States.active,
        ]

    @staticmethod
    def mpijob_state_to_run_state(mpijob_state):
        if mpijob_state not in MPIJobV1Alpha1States.all():
            raise ValueError(f"Invalid MPI job state: {mpijob_state}")
        return {
            MPIJobV1Alpha1States.succeeded: RunStates.completed,
            SparkApplicationStates.failed: RunStates.error,
            MPIJobV1Alpha1States.active: RunStates.running,
        }[mpijob_state]


class MPIJobV1CleanPodPolicies:
    """
    https://github.com/kubeflow/common/blob/master/pkg/apis/common/v1/types.go#L137
    """

    all = "All"
    running = "Running"
    none = "None"
    undefined = ""

    @staticmethod
    def default():
        return MPIJobV1CleanPodPolicies.all


class NuclioIngressAddTemplatedIngressModes:
    always = "always"
    never = "never"
    on_cluster_ip = "onClusterIP"


class FunctionEnvironmentVariables:
    _env_prefix = "MLRUN_"
    auth_session = f"{_env_prefix}AUTH_SESSION"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle
from typing import Optional, Union

import cloudpickle
import lightgbm as lgb

import mlrun

from .._common import without_mlrun_interface
from .._ml_common import MLModelHandler
from .mlrun_interfaces.booster_mlrun_interface import LGBMBoosterMLRunInterface
from .mlrun_interfaces.model_mlrun_interface import LGBMModelMLRunInterface
from .utils import LGBMTypes


class LGBMModelHandler(MLModelHandler):
    """
    Class for handling a XGBoost model, enabling loading and saving it during runs.
    """

    # Framework name:
    FRAMEWORK_NAME = "lightgbm"

    class ModelFormats:
        """
        Model formats to pass to the 'LGBMModelHandler' for loading and saving LightGBM models.
        """

        PKL = "pkl"
        TXT = "txt"

    class _LabelKeys:
        """
        Required labels keys to log with the model.
        """

        MODEL_FORMAT = "model_format"

    def __init__(
        self,
        model_name: Optional[str] = None,
        model_path: Optional[str] = None,
        model: LGBMTypes.ModelType = None,
        modules_map: Optional[
            Union[dict[str, Union[None, str, list[str]]], str]
        ] = None,
        custom_objects_map: Optional[
            Union[dict[str, Union[str, list[str]]], str]
        ] = None,
        custom_objects_directory: Optional[str] = None,
        context: mlrun.MLClientCtx = None,
        model_format: str = ModelFormats.PKL,
        **kwargs,
    ):
        """
        Initialize the handler. The model can be set here, so it won't require loading. Note you must provide at least
        one of 'model' and 'model_path'. If a model is not given, the files in the model path will be collected
        automatically to be ready for loading.

        :param model:                    Model to handle or None in case a loading parameters were supplied.
        :param model_path:               Path to the directory with the model files. Can be passed as a model object
                                         path in the following format:
                                         'store://models/<PROJECT_NAME>/<MODEL_NAME>:<VERSION>'
        :param model_name:               The model name for saving and logging the model:
                                         * Mandatory for loading the model from a local path.
                                         * If given a logged model (store model path) it will be read from the artifact.
                                         * If given a loaded model object and the model name is None, the name will be
                                           set to the model's object class.
        :param modules_map:              A dictionary of all the modules required for loading the model. Each key is a
                                         path to a module and its value is the object name to import from it. All the
                                         modules will be imported globally. If multiple objects needed to be imported
                                         from the same module a list can be given. The map can be passed as a path to a
                                         json file as well. For example:

                                         .. code-block:: python

                                             {
                                                 "module1": None,  # import module1
                                                 "module2": ["func1", "func2"],  # from module2 import func1, func2
                                                 "module3.sub_module": "func3",  # from module3.sub_module import func3
                                             }

                                         If the model path given is of a store object, the modules map will be read from
                                         the logged modules map artifact of the model.
        :param custom_objects_map:       A dictionary of all the custom objects required for loading the model. Each key
                                         is a path to a python file and its value is the custom object name to import
                                         from it. If multiple objects needed to be imported from the same py file a list
                                         can be given. The map can be passed as a path to a json file as well. For
                                         example:

                                         .. code-block:: python

                                             {
                                                 "/.../custom_model.py": "MyModel",
                                                 "/.../custom_objects.py": ["object1", "object2"],
                                             }

                                         All the paths will be accessed from the given 'custom_objects_directory',
                                         meaning each py file will be read from 'custom_objects_directory/<MAP VALUE>'.
                                         If the model path given is of a store object, the custom objects map will be
                                         read from the logged custom object map artifact of the model.
                                         Notice: The custom objects will be imported in the order they came in this
                                         dictionary (or json). If a custom object is depended on another, make sure to
                                         put it below the one it relies on.
        :param custom_objects_directory: Path to the directory with all the python files required for the custom
                                         objects. Can be passed as a zip file as well (will be extracted during the run
                                         before loading the model). If the model path given is of a store object, the
                                         custom objects files will be read from the logged custom object artifact of the
                                         model.
        :param context:                  MLRun context to work with for logging the model.
        :param model_format:             The format to use for saving and loading the model. Should be passed as a
                                         member of the class 'LGBMModelHandler.ModelFormats'. Default:
                                         'LGBMModelHandler.ModelFormats.PKL'.

        :raise MLRunInvalidArgumentError: In case one of the given parameters are invalid.
        """
        # Validate given format:
        if model_format not in [
            LGBMModelHandler.ModelFormats.PKL,
            LGBMModelHandler.ModelFormats.TXT,
        ]:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unrecognized model format: '{model_format}'. Please use one of the class members of "
                "'TFKerasModelHandler.ModelFormats'"
            )

        # Store the configuration:
        self._model_format = model_format

        # Set up the base handler class:
        super().__init__(
            model=model,
            model_path=model_path,
            model_name=model_name,
            modules_map=modules_map,
            custom_objects_map=custom_objects_map,
            custom_objects_directory=custom_objects_directory,
            context=context,
            **kwargs,
        )

    def set_labels(
        self,
        to_add: Optional[dict[str, Union[str, int, float]]] = None,
        to_remove: Optional[list[str]] = None,
    ):
        """
        Update the labels dictionary of this model artifact. There are required labels that cannot be edited or removed.

        :param to_add:    The labels to add.
        :param to_remove: A list of labels keys to remove.
        """
        # Update the user's labels:
        super().set_labels(to_add=to_add, to_remove=to_remove)

        # Set the required labels:
        self._labels[self._LabelKeys.MODEL_FORMAT] = self._model_format

    def _collect_files_from_local_path(self):
        """
        If the model path given is of a local path, search for the needed model files and collect them into this handler
        for later loading the model.

        :raise MLRunNotFoundError: If the model file was not found.
        """
        self._model_file = os.path.join(
            self._model_path, f"{self._model_name}.{self._model_format}"
        )
        if not os.path.exists(self._model_file):
            raise mlrun.errors.MLRunNotFoundError(
                f"The model file '{self._model_name}.{self._model_format}' was not found within the given "
                f"'model_path': '{self._model_path}'"
            )

    def save(self, output_path: Optional[str] = None, **kwargs):
        """
        Save the handled model at the given output path. If a MLRun context is available, the saved model files will be
        logged and returned as artifacts.

        :param output_path: The full path to the directory to save the handled model at. If not given, the context
                            stored will be used to save the model in the default artifacts location.

        :return The saved model additional artifacts (if needed) dictionary if context is available and None otherwise.
        """
        super().save(output_path=output_path)

        if isinstance(self._model, lgb.LGBMModel):
            return self._save_lgbmmodel()
        return self._save_booster()

    def load(self, **kwargs):
        """
        Load the specified model in this handler. Additional parameters for the class initializer can be passed via the
        kwargs dictionary.
        """
        super().load()

        # ModelFormats.PKL - Load from a pkl file:
        if self._model_format == LGBMModelHandler.ModelFormats.PKL:
            with open(self._model_file, "rb") as pickle_file:
                self._model = pickle.load(pickle_file)
            return

        # ModelFormats.TXT:
        self._model = lgb.Booster(model_file=self._model_file)

    def to_onnx(
        self,
        model_name: Optional[str] = None,
        optimize: bool = True,
        input_sample: LGBMTypes.DatasetType = None,
        log: Optional[bool] = None,
    ):
        """
        Convert the model in this handler to an ONNX model. The inputs names are optional, they do not change the
        semantics of the model, it is only for readability.

        :param model_name:          The name to give to the converted ONNX model. If not given the default name will be
                                    the stored model name with the suffix '_onnx'.
        :param optimize:            Whether to optimize the ONNX model using 'onnxoptimizer' before saving the model.
                                    Default: True.
        :param input_sample:        An inputs sample with the names and data types of the inputs of the model.
        :param log:                 In order to log the ONNX model, pass True. If None, the model will be logged if this
                                    handler has a MLRun context set. Default: None.

        :return: The converted ONNX model (onnx.ModelProto).

        :raise MLRunMissingDependencyError: If some ONNX packages are missing.
        """
        # Import onnx related modules:
        try:
            pass
            # import onnxmltools

            # from mlrun.frameworks.onnx import ONNXModelHandler
        except ModuleNotFoundError:
            raise mlrun.errors.MLRunMissingDependencyError(
                "ONNX conversion requires additional packages to be installed. "
                "Please run 'pip install mlrun[lgbm]' to install MLRun's XGBoost package."
            )

        raise NotImplementedError  # TODO: Finish ONNX conversion

    @without_mlrun_interface(interface=LGBMModelMLRunInterface)
    def _save_lgbmmodel(self):
        """
        Save the model handled in this handler (must be a type of `lgb.LGBMModel`).
        """
        # ModelFormats.PKL:
        if self._model_format == LGBMModelHandler.ModelFormats.PKL:
            self._model_file = f"{self._model_name}.pkl"
            with open(self._model_file, "wb") as pickle_file:
                cloudpickle.dump(self._model, pickle_file)
            return

        # Other model formats require saving directly from the booster object:
        return self._save_booster(booster=self.model.booster_)

    @without_mlrun_interface(interface=LGBMBoosterMLRunInterface)
    def _save_booster(self, booster: lgb.Booster = None):
        """
        Save the booster object in the format saved in this handler. If not booster passed, the model handled in this
        handler will be used (must be a type of `lgb.Booster` object).

        :param booster: The booster object to save.
        """
        # Set the model as the booster to save in case no booster was provided:
        if booster is None:
            booster = self._model

        # ModelFormats.PICKLE:
        if self._model_format == LGBMModelHandler.ModelFormats.PKL:
            self._model_file = f"{self._model_name}.pkl"
            with open(self._model_file, "wb") as pickle_file:
                cloudpickle.dump(booster, pickle_file)
            return

        # ModelFormats.TXT:
        self._model_file = f"{self._model_name}.txt"
        booster.save_model(self._model_file)
